import logging

import numpy as np
from sklearn.ensemble import RandomForestRegressor, BaggingRegressor
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, ConstantKernel, Matern, WhiteKernel, ExpSineSquared, RationalQuadratic
from sklearn.metrics import mean_squared_error
from sklearn.neural_network import MLPRegressor
from sklearn.svm import SVR
from sklearn.tree import DecisionTreeRegressor

logger = logging.getLogger("GA_algo")


# TODO: Add fitness type
def set_surrogate_fitness(value, fitness_type="avg_coherence_score"):
    npmis = {
        "npmi_50": None,
        "npmi_15": None,
        "npmi_25": None,
        "npmi_50_list": None,
    }
    scores_dict = {
        fitness_type: value,
        "perplexityScore": None,
        "backgroundTokensRatioScore": None,
        "contrast": None,
        "purity": None,
        "kernelSize": None,
        "npmi_50_list": [None],  # npmi_values_50_list,
        "npmi_50": None,
        "sparsity_phi": None,
        "sparsity_theta": None,
        "topic_significance_uni": None,
        "topic_significance_vacuous": None,
        "topic_significance_back": None,
        "switchP_list": [None],
        "switchP": None,
        "all_topics": None,
        # **coherence_scores,
        **npmis,
    }
    return scores_dict


class Surrogate:
    def __init__(self, surrogate_name, **kwargs):
        self.name = surrogate_name
        self.kwargs = kwargs
        self.surrogate = None
        self.br_n_estimators = None
        self.br_n_jobs = None
        self.gpr_kernel = None

    def create(self):
        kernel = self.kwargs["gpr_kernel"]
        del self.kwargs["gpr_kernel"]
        gpr_alpha = self.kwargs["gpr_alpha"]
        del self.kwargs["gpr_alpha"]
        normalize_y = self.kwargs["normalize_y"]
        del self.kwargs["normalize_y"]

        if self.name == "random-forest-regressor":
            self.surrogate = RandomForestRegressor(**self.kwargs)
        elif self.name == "mlp-regressor":
            if not self.br_n_estimators:
                self.br_n_estimators = self.kwargs["br_n_estimators"]
                del self.kwargs["br_n_estimators"]
                self.br_n_jobs = self.kwargs["n_jobs"]
                del self.kwargs["n_jobs"]
                self.kwargs["alpha"] = self.kwargs["mlp_alpha"]
                del self.kwargs["mlp_alpha"]
            self.surrogate = BaggingRegressor(
                base_estimator=MLPRegressor(**self.kwargs),
                n_estimators=self.br_n_estimators,
                n_jobs=self.br_n_jobs,
            )
        elif self.name == "GPR":  # tune ??
            if not self.gpr_kernel:
                if kernel == "RBF":
                    self.gpr_kernel = 1.0 * RBF(1.0)
                elif kernel == "RBFwithConstant":
                    self.gpr_kernel = 1.0 * RBF(1.0) + ConstantKernel()
                elif kernel == "Matern":
                    self.gpr_kernel = 1.0 * Matern(1.0)
                elif kernel == "WhiteKernel":
                    self.gpr_kernel = 1.0 * WhiteKernel(1.0)
                elif kernel == "ExpSineSquared":
                    self.gpr_kernel = ExpSineSquared()
                elif kernel == "RationalQuadratic":
                    self.gpr_kernel = RationalQuadratic(1.0)
                self.kwargs["kernel"] = self.gpr_kernel
                self.kwargs["alpha"] = gpr_alpha
                self.kwargs["normalize_y"] = normalize_y
            self.surrogate = GaussianProcessRegressor(**self.kwargs)
        elif self.name == "decision-tree-regressor":
            try:
                if self.kwargs["max_depth"] == 0:
                    self.kwargs["max_depth"] = None
            except KeyError:
                logger.error("No max_depth")
            self.surrogate = DecisionTreeRegressor(**self.kwargs)
        elif self.name == "SVR":
            self.surrogate = SVR(**self.kwargs)
        # else:
        #     raise Exception('Undefined surr')

    def fit(self, X, y):
        logger.debug(f"X: {X}, y: {y}")
        self.create()
        self.surrogate.fit(X, y)

    def score(self, X, y):
        r_2 = self.surrogate.score(X, y)
        y_pred = self.surrogate.predict(X)
        mse = mean_squared_error(y, y_pred)
        rmse = np.sqrt(mse)
        return r_2, mse, rmse

    def predict(self, X):
        m = self.surrogate.predict(X)
        return m


def get_prediction_uncertanty(model, X, surrogate_name, percentile=90):
    interval_len = []
    if surrogate_name == "random-forest-regressor":
        for x in range(len(X)):
            preds = []
            for pred in model.estimators_:
                prediction = pred.predict(np.array(X[x]).reshape(1, -1))
                preds.append(prediction[0])
            err_down = np.percentile(preds, (100 - percentile) / 2.0)
            err_up = np.percentile(preds, 100 - (100 - percentile) / 2.0)
            interval_len.append(err_up - err_down)
    elif surrogate_name == "GPR":
        y_hat, y_sigma = model.predict(X, return_std=True)
        interval_len = list(y_sigma)
    elif surrogate_name == "decision-tree-regressor":
        raise NotImplementedError
    return interval_len



from typing import Tuple, Callable

from autotm.algorithms_for_tuning.genetic_algorithm.crossover import crossover_one_point
from autotm.pipeline import *


# util methods

def random_action(probability: float):
    """
    Bernoulli's distribution.
    :param probability: True probability
    :return: True with given probability, or False otherwise
    """
    return random.random() < probability


def create_or_default(create: Callable, probability: float, default_value):
    """
    Create new value with given probability or use the default one otherwise.
    """
    return create() if random_action(probability) else default_value


def replace_values(original: List, new: List):
    """
    Replace values in a list with values from another list.
    """
    original.clear()
    for e in new:
        original.append(e)


def sort_by_values(keys: List, values: List[float]):
    """
    Sort given keys by associated values.
    """
    data = list(zip(values, keys))
    data.sort(reverse=True)
    sorted_values, sorted_keys = zip(*data)

    replace_values(keys, sorted_keys)
    replace_values(values, sorted_values)


def sort_by_value(keys: List, value_gen: Callable) -> List[float]:
    """
    Sort given keys by values generated via passed lambda and return the values.
    """
    values = list(map(value_gen, keys))
    sort_by_values(keys, values)
    return values


def remove_all(elements, elements_to_remove):
    for e in elements_to_remove:
        if e in elements:
            elements.remove(e)


def single_point_crossover(list1: List, list2: List) -> Tuple[List, List]:
    while True:
        i1 = random.randint(0, len(list1))
        i2 = random.randint(0, len(list2))
        if (i1 == 0 and i2 == 0) or (i1 == len(list1) and i2 == len(list2)):
            continue
        break

    return list1[:i1] + list2[i2:], list2[:i2] + list1[i1:]


def crossover_stage(stage1: Stage, stage2: Stage) -> Tuple[Stage, Stage]:
    if stage1.stage_type != stage2.stage_type:
        raise ValueError("Cannot crossover stages with different types")
    values1, values2 = crossover_one_point(stage1.values, stage2.values, elem_cross_prob=0.2)
    return Stage(stage_type=stage1.stage_type, values=values1), Stage(stage_type=stage2.stage_type, values=values2)


def crossover_pipelines(pipeline1: Pipeline, pipeline2: Pipeline) -> List[Pipeline]:
    """
    Single-point crossover of two pipelines.
    Returns non-empty pipelines after crossover.
    """

    stages1, stages2 = single_point_crossover(pipeline1.stages, pipeline2.stages)
    values1, values2 = crossover_stage(pipeline1.required_params, pipeline2.required_params)

    child1 = Pipeline(stages=stages1, required_params=values1)
    child2 = Pipeline(stages=stages2, required_params=values2)

    return [child for child in [child1, child2] if len(child.stages) > 0]


def mutate_stage(stage: Stage, mutation_probability: float) -> Stage:
    """
    Mutate parameters of a single stage.
    :param stage: stage to mutate
    :param mutation_probability: Probability of single parameter mutation
    :return: a mutated copy of the stage
    """
    mutated_values = [create_or_default(param.create_value, mutation_probability, value) for param, value
                      in zip(stage.stage_type.params, stage.values)]
    return Stage(stage_type=stage.stage_type, values=mutated_values)


def mutate_required_params(pipeline: Pipeline, mutation_probability: float) -> Pipeline:
    """
    Mutate required parameters of the pipeline.
    :param pipeline: pipeline to mutate
    :param mutation_probability: Probability of single parameter mutation
    :return: a mutated copy of the pipeline
    """
    mutated_params = mutate_stage(pipeline.required_params, mutation_probability)
    return Pipeline(stages=pipeline.stages, required_params=mutated_params)


def mutate_stages(pipeline: Pipeline, mutation_probability: float) -> Pipeline:
    """
    Mutate parameters of a single stage from the pipeline.
    :param pipeline: pipeline to mutate
    :param mutation_probability: Probability of single parameter mutation
    :return: a mutated copy of the pipeline
    """
    i = pipeline.random_stage_index()
    mutated_stage = mutate_stage(pipeline.stages[i], mutation_probability)
    return Pipeline(stages=pipeline.stages[:i] + [mutated_stage] + pipeline.stages[i + 1:],
                    required_params=pipeline.required_params)


def remove_stage(pipeline: Pipeline):
    """
    Remove random stage from the pipeline.
    Can be applied only to pipelines with size more than 1.
    :param pipeline: pipeline to mutate
    :return: a mutated copy of the pipeline
    """
    if len(pipeline.stages) <= 1:
        return pipeline
    deletion_index = pipeline.random_stage_index()
    return Pipeline(stages=pipeline.stages[:deletion_index] + pipeline.stages[deletion_index + 1:],
                    required_params=pipeline.required_params)


def add_stage(pipeline: Pipeline, stage_types: List[StageType]):
    """
    Add random stage to the pipeline.
    :param pipeline: pipeline to mutate
    :param stage_types: stage types that can be used as a new stage
    :return: a mutated copy of the pipeline
    """
    stage = create_stage(random.choice(stage_types))
    insertion_index = pipeline.random_stage_index(with_last=True)
    return Pipeline(stages=pipeline.stages[:insertion_index] + [stage] + pipeline.stages[insertion_index:],
                    required_params=pipeline.required_params)


def swap_stages(pipeline: Pipeline):
    """
    Swap to random stages in the pipeline.
    Can be applied only to pipelines with size more than 1.
    :param pipeline: pipeline to mutate
    :return: a mutated copy of the pipeline
    """
    if len(pipeline.stages) <= 1:
        return pipeline
    i1 = pipeline.random_stage_index()
    i2 = pipeline.random_stage_index()
    while i1 == i2:
        i2 = pipeline.random_stage_index()

    stages = [stage for stage in pipeline.stages]
    stages[i1], stages[i2] = stages[i2], stages[i1]
    return Pipeline(stages=stages, required_params=pipeline.required_params)


def mutate_pipeline(pipeline: Pipeline, stage_types: List[StageType], stage_mutation_probability: float) -> Pipeline:
    """
    Apply one mutation to the pipeline.
    :param pipeline: pipeline to mutate
    :param stage_types: stage types that can be used as a new stage
    :param stage_mutation_probability: Probability of single parameter mutation
    :return: a mutated copy of the pipeline
    """
    mutations = [lambda p: add_stage(p, stage_types),
                 lambda p: mutate_stages(p, stage_mutation_probability),
                 ]
    weights = [1, 7]
    if len(pipeline.stages) > 1:
        mutations.append(remove_stage)
        weights.append(1)
        mutations.append(swap_stages)
        weights.append(1)
    if len(pipeline.required_params.stage_type.params) > 0:
        mutations.append(lambda p: mutate_required_params(p, stage_mutation_probability))
        weights.append(3)
    mutation = random.choices(mutations, weights=weights)[0]
    return mutation(pipeline)


def create_pipeline(stage_types: List[StageType],
                    pipeline_size_distribution: Callable[[], int],
                    required_stage_type: StageType) -> Pipeline:
    size = pipeline_size_distribution()
    return Pipeline(stages=list(map(create_stage, random.choices(stage_types, k=size))),
                    required_params=create_stage(required_stage_type))


def create_initial_population(stage_types: List[StageType],
                              population_size: int,
                              pipeline_size_distribution: Callable[[], int],
                              required_stage_type: StageType) -> List[Pipeline]:
    return [create_pipeline(stage_types, pipeline_size_distribution, required_stage_type)
            for _ in range(population_size)]


def tournament_selection(population: List[Pipeline], tournament_size: int) -> List[Pipeline]:
    tournament_contestants = random.sample(range(len(population)), tournament_size)
    tournament_contestants.sort()
    w1, w2 = tournament_contestants[:2]
    return [population[w1], population[w2]]


def genetic_algorithm(population: List[Pipeline],
                      generations: int,
                      stage_types: List[StageType],
                      fitness: Callable[[Pipeline], float],
                      crossover_probability: float,
                      mutation_probability: float,
                      stage_mutation_probability: float,
                      generation_callback: Callable[[List[Pipeline], List[float]], None] = None,
                      tournament_size=5) -> Tuple[List[Pipeline], List[float]]:
    population_size = len(population)
    fitness_values = sort_by_value(population, fitness)

    for generation in range(generations):
        new_population = []

        while len(new_population) < population_size:
            parent1, parent2 = tournament_selection(population, tournament_size)
            children = create_or_default(lambda: crossover_pipelines(parent1, parent2),
                                         crossover_probability,
                                         [parent1, parent2])
            children = [create_or_default(lambda: mutate_pipeline(child, stage_types, stage_mutation_probability),
                                          mutation_probability,
                                          child) for child in children]
            remove_all(children, [parent1, parent2])
            new_population += children

        new_fitness_values = sort_by_value(new_population, fitness)

        population += new_population
        fitness_values += new_fitness_values

        sort_by_values(population, fitness_values)

        if generation_callback is not None:
            generation_callback(population, fitness_values)

        population = population[:population_size]
        fitness_values = fitness_values[:population_size]

    return population, fitness_values



import os

import yaml


def main():
    os.environ['AUTOTM_COMPONENT'] = 'worker'
    os.environ['AUTOTM_EXEC_MODE'] = 'cluster'

    from autotm.fitness.cluster_tasks import make_celery_app
    from autotm.fitness.tm import TopicModelFactory

    if "DATASETS_CONFIG" in os.environ:
        with open(os.environ["DATASETS_CONFIG"], "r") as f:
            config = yaml.safe_load(f)
            dataset_settings = config["datasets"]
    else:
        dataset_settings = None

    TopicModelFactory.init_factory_settings(
        num_processors=os.getenv("NUM_PROCESSORS", None),
        dataset_settings=dataset_settings
    )

    celery_app = make_celery_app()
    celery_app.worker_main()


if __name__ == '__main__':
    main()



import random
import numpy as np
from typing import List, Callable


def crossover_pmx(parent_1: List[float], parent_2: List[float], **kwargs) -> List[List[float]]:
    """
    Pmx crossover

    Exchange chromosome parts

    Parameters
    ----------
    parent_1: List[float]
        The first individual to be processed
    parent_2: List[float]
        The second individual to be processed

    Returns
    ----------
    Updated individuals with exchanged chromosome parts
    """
    points_num = len(parent_1)
    while True:
        cut_ix = np.random.choice(points_num + 1, 2, replace=False)
        min_ix = np.min(cut_ix)
        max_ix = np.max(cut_ix)
        part = parent_1[min_ix:max_ix]
        if len(part) != len(parent_1):
            break
    parent_1[min_ix:max_ix] = parent_2[min_ix:max_ix]
    parent_2[min_ix:max_ix] = part
    return [parent_1, parent_2]


# discrete crossover
def crossover_one_point(parent_1: List[float], parent_2: List[float], **kwargs) -> List[List[float]]:
    """
    One-point crossover

    Exchange points between chromosomes

    Parameters
    ----------
    parent_1: List[float]
        The first individual to be processed
    parent_2: List[float]
        The second individual to be processed

    Returns
    ----------
    Updated individuals with exchanged chromosome parts
    """
    elem_cross_prob = kwargs["elem_cross_prob"]
    for i in range(len(parent_1)):
        # removed mutation preservation
        if random.random() < elem_cross_prob:
            parent_1[i], parent_2[i] = parent_2[i], parent_1[i]
    return [parent_1, parent_2]


def crossover_blend_new(parent_1: List[float], parent_2: List[float], **kwargs) -> List[List[float]]:
    """
    Blend crossover

    Making combination of parents solution with coefficient

    Parameters
    ----------
    parent_1: List[float]
        The first individual to be processed
    parent_2: List[float]
        The second individual to be processed
    alpha: float
        Blending coefficient

    Returns
    ----------
    Updated individuals with exchanged chromosome parts
    """
    alpha = kwargs["alpha"]
    child_1 = []
    child_2 = []
    u = random.random()
    gamma = (1.0 + 2.0 * alpha) * u - alpha  # fixed (1. + 2. * alpha) * u - alpha
    for i in range(len(parent_1)):
        child_1.append((1.0 - gamma) * parent_1[i] + gamma * parent_2[i])
        child_2.append(gamma * parent_1[i] + (1.0 - gamma) * parent_2[i])

    # TODO: reconsider this
    child_1[12:15] = parent_1[12:15]
    child_2[12:15] = parent_2[12:15]
    return [child_1, child_2]


def crossover_blend(parent_1: List[float], parent_2: List[float], **kwargs) -> List[List[float]]:
    """
    Blend crossover

    Making combination of parents solution with coefficient

    Parameters
    ----------
    parent_1: List[float]
        The first individual to be processed
    parent_2: List[float]
        The second individual to be processed
    alpha: float
        Blending coefficient

    Returns
    ----------
    Updated individuals with exchanged chromosome parts
    """
    alpha = kwargs["alpha"]
    child = []
    u = random.random()
    gamma = (1 - 2 * alpha) * u - alpha
    for i in range(len(parent_1)):
        child.append((1 - gamma) * parent_1[i] + gamma * parent_2[i])
    if random.random() > 0.5:
        child[12:15] = parent_1[12:15]
    else:
        child[12:15] = parent_2[12:15]
    return [child]


def crossover(crossover_type: str = "crossover_one_point") -> Callable:
    """
    Crossover function

    Parameters
    ----------
    crossover_type : str, default="crossover_one_point"
        Crossover to be used in the genetic algorithm
    """
    if crossover_type == "crossover_pmx":
        return crossover_pmx
    if crossover_type == "crossover_one_point":
        return crossover_one_point
    if crossover_type == "blend_crossover":
        return crossover_blend



# source: https://github.com/machine-intelligence-laboratory/TopicNet/blob/88963c16c65b90789739419ec1697843c9a97129/topicnet/cooking_machine/models/base_score.py#L12

import dill

from typing import (
    Any,
    Callable,
    Dict,
)

# from autotm.fitness import scores as tn_scores


class BaseScore:
    """
    Base Class to construct custom score functions.
    """

    _PRECOMPUTED_DATA_PARAMETER_NAME = "precomputed_data"

    # TODO: name should not be optional
    def __init__(
        self, name: str = None, should_compute: Callable[[int], bool] or bool = None
    ):
        """
        Parameters
        ----------
        name
            Name of the score
        should_compute
            Function which decides whether the score should be computed
            on the current fit iteration or not.
            If `should_compute` is `None`, then score is going to be computed on every iteration.
            At the same time, whatever function one defines,
            score is always computed on the last fit iteration.
            This is done for two reasons.
            Firstly, so that the score is always computed at least once during `model._fit()`.
            Secondly, so that `experiment.select()` works correctly.
            The parameter `should_compute` might be helpful
            if the score is slow but one still needs
            to get the dependence of the score on iteration
            (for the described case, one may compute the score
            on every even iteration or somehow else).
            However, be aware that if `should_compute` is used for some model's scores,
            then the scores may have different number of values in `model.scores`!
            Number of score values is the number of times the scores was calculated;
            first value corresponds to the first fit iteration
            which passed `should_compute` etc.
            There are a couple of things also worth noting.
            Fit iteration numbering starts from zero.
            And every new `model._fit()` call is a new range of fit iterations.
        Examples
        --------
        Scores created below are unworkable (as BaseScore has no `call` method inplemented).
        These are just the examples of how one can create a score and set some of its parameters.
        Scores to be computed on every iteration:
        >>> score = BaseScore()
        >>> score = BaseScore(should_compute=BaseScore.compute_always)
        >>> score = BaseScore(should_compute=lambda i: True)
        >>> score = BaseScore(should_compute=True)
        Scores to be computed only on the last iteration:
        >>> score = BaseScore(should_compute=BaseScore.compute_on_last)
        >>> score = BaseScore(should_compute=lambda i: False)
        >>> score = BaseScore(should_compute=False)
        Score to be computed only on even iterations:
        >>> score = BaseScore(should_compute=lambda i: i % 2 == 0)
        """
        self._name = name

        if should_compute is None:
            should_compute = self.compute_always
        elif should_compute is True:
            should_compute = self.compute_always
        elif should_compute is False:
            should_compute = self.compute_on_last
        elif not isinstance(should_compute, type(lambda: None)):
            raise TypeError(
                f"Unknown type of `should_compute`: {type(should_compute)}!"
            )
        else:
            pass

        self._should_compute = should_compute
        self.value = []

        # if not hasattr(tn_scores, self.__class__.__name__):
        #     setattr(tn_scores, self.__class__.__name__, self.__class__)

    @staticmethod
    def compute_always(fit_iteration: int) -> bool:
        return True

    @staticmethod
    def compute_on_last(fit_iteration: int) -> bool:
        return False

    def __repr__(self):
        return f"{self.__class__.__name__}"

    def save(self, path):
        with open(path, "wb") as f:
            dill.dump(self, f)

    @classmethod
    def load(cls, path):
        with open(path, "rb") as f:
            score = dill.load(f)

        return score

    def update(self, score):
        """
        Parameters
        ----------
        score : float
            score value
        Returns
        -------
        """
        known_errors = (ValueError, TypeError)

        try:
            score = float(score)
        except known_errors:
            raise ValueError(f"Score call should return float but not {score}")

        self.value.append(score)

    def call(self, model, precomputed_data: Dict[str, Any] = None):
        """
        Call to custom score function.
        Parameters
        ----------
        model : TopicModel
            a TopicNet model inherited from BaseModel
        precomputed_data
            Data which scores may share between each other during *one fit iteration*.
            For example, if the model has several scores of the same score class,
            and there is a heavy time consuming computation inside this score class,
            it may be useful to perform the calculations *only once*, for one score instance,
            and then make the result visible for all other scores that might need it.
        Returns
        -------
        float
            score
        Notes
        -----
        Higher score not necessarily should correspond to better model.
        It is up to user to decide what the meaning is behind the score,
        and then use this logic in query in Experiment's `select()` method.
        If one need ARTM model for score (not TopicNet one), it is available as model._model
        When creating a custom score class,
        it is recommended to use `**kwargs` in the score's `call` method,
        so that all `BaseScore` optional parameters are also available
        in its successor score classes.
        Examples
        --------
        Score which uses `precomputed_data`:
        >>> import time
        ...
        >>> class NewScore(BaseScore):
        ...     def __init__(self, name: str, multiplier: float):
        ...         super().__init__(name=name)
        ...
        ...         self._multiplier = multiplier
        ...         self._heavy_value_name = 'time_consuming_value_name'
        ...
        ...     def call(self, model, precomputed_data = None):
        ...         if precomputed_data is None:
        ...             # Parameter `precomputed_data` is optional in BaseScore
        ...             # So this case also should be supported
        ...             heavy_value = self._compute_heavy(model)
        ...         elif self._heavy_value_name in precomputed_data:
        ...             # This is going to be fast
        ...             heavy_value = precomputed_data[self._heavy_value_name]
        ...         else:
        ...             # This is slow (but only one such call!)
        ...             heavy_value = self._compute_heavy(model)
        ...             precomputed_data[self._heavy_value_name] = heavy_value
        ...
        ...         return heavy_value * self._multiplier
        ...
        ...     def _compute_heavy(self, model):
        ...         time.sleep(100)  # just for demonstration
        ...
        ...         return 0
        """
        raise NotImplementedError("Define your score here")



import logging
import os
import time
import warnings

import numpy as np
import pandas as pd
import plotly.express as px
from scipy.spatial import distance

GENERATION_COL = "generation"
FITNESS_COL = "fitness"
FITNESS_DIFF_COL = "fitness_diff"
PARAMS_DIST_COL = "params_dist"

logger = logging.getLogger(__name__)


class MetricsCollector:
    def __init__(
        self, dataset, n_specific_topics, experiment_id=None, save_path="./metrics/"
    ):
        """

        :param dataset: dataset name
        :param n_specific_topics: number of specific (main) topics
        :param experiment_id: unique id of the experiment, by default time.time()
        :param save_path: folder where to store the results of experiments, be default ./metrics/
        """
        if not experiment_id:
            experiment_id = str(int(time.time()))
        self.save_path = os.path.join(save_path)
        self.save_fname = f"{experiment_id}_{dataset}_{n_specific_topics}"
        self.dict_of_population = {}
        self.mutation_changes = {}
        self.crossover_changes = {}
        self.num_generations = 0
        self.metric_df = None
        self.mutation_df = None
        self.crossover_df = None

    def save_mutation(
        self,
        generation: int,
        original_params: list,
        mutated_params: list,
        original_fitness: float,
        mutated_fitness: float,
    ):
        # excluding mutation parameters
        eucledian_distance = distance.euclidean(
            original_params[:11] + [original_params[15]],
            mutated_params[:11] + [mutated_params[15]],
        )
        fitness_diff = mutated_fitness - original_fitness

        if f"gen_{generation}" in self.mutation_changes:
            self.mutation_changes[f"gen_{generation}"]["params_dist"].append(
                eucledian_distance
            )
            self.mutation_changes[f"gen_{generation}"]["fitness_diff"].append(
                fitness_diff
            )
            self.mutation_changes[f"gen_{generation}"]["original_params"].append(
                original_params[:12] + [original_params[15]]
            )
            self.mutation_changes[f"gen_{generation}"]["mutated_params"].append(
                mutated_params[:12] + [mutated_params[15]]
            )
        else:
            self.mutation_changes[f"gen_{generation}"] = {
                "params_dist": [eucledian_distance],
                "fitness_diff": [fitness_diff],
                "original_params": [original_params[:11] + [original_params[15]]],
                "mutated_params": [mutated_params[:11] + [mutated_params[15]]],
            }

    def save_crossover(
        self,
        generation: int,
        parent_1: list,
        parent_2: list,
        child_1: list,
        parent_1_fitness: float,
        parent_2_fitness: float,
        child_1_fitness: float,
        child_2: list = None,
        child_2_fitness: list = None,
    ):
        """

        :param generation: generation number
        :param parent_1: parameters of the first parent
        :param parent_2: parameters of the second parent
        :param child_1: parameters of the first child
        :param parent_1_fitness: fitness of the first parent
        :param parent_2_fitness:  fitness of the second parent
        :param child_1_fitness: fitness of the firs child
        :param child_2: parameters of the second child if exists
        :param child_2_fitness: fitness of the second child if exists
        :return:
        """
        if f"gen_{generation}" in self.crossover_changes:
            self.crossover_changes[f"gen_{generation}"]["parent_1_params"].append(
                parent_1
            )
            self.crossover_changes[f"gen_{generation}"]["parent_2_params"].append(
                parent_2
            )
            self.crossover_changes[f"gen_{generation}"]["child_1_params"].append(
                child_1
            )
            self.crossover_changes[f"gen_{generation}"]["parent_1_fitness"].append(
                parent_1_fitness
            )
            self.crossover_changes[f"gen_{generation}"]["parent_2_fitness"].append(
                parent_2_fitness
            )
            self.crossover_changes[f"gen_{generation}"]["child_1_fitness"].append(
                child_1_fitness
            )
            if child_2 is not None:
                self.crossover_changes[f"gen_{generation}"]["child_2_params"].append(
                    child_2
                )
                self.crossover_changes[f"gen_{generation}"]["child_2_fitness"].append(
                    child_2_fitness
                )
        else:
            self.crossover_changes[f"gen_{generation}"] = {
                "parent_1_params": [parent_1],
                "parent_2_params": [parent_2],
                "child_1_params": [child_1],
                "parent_1_fitness": [parent_1_fitness],
                "parent_2_fitness": [parent_2_fitness],
                "child_1_fitness": [child_1_fitness],
            }
            if child_2 is not None:
                self.crossover_changes[f"gen_{generation}"]["child_2_params"] = [
                    child_2
                ]
                self.crossover_changes[f"gen_{generation}"]["child_2_fitness"] = [
                    child_2_fitness
                ]

    def save_fitness(self, generation: int, params: list, fitness: float):
        """

        :param generation: id of the generation
        :param params: hyperparameters of the individ
        :param fitness: fitness value
        :return:
        """

        if generation > self.num_generations:
            self.num_generations = generation

        if f"gen_{generation}" in self.dict_of_population:
            self.dict_of_population[f"gen_{generation}"]["fitness"].append(fitness)
            self.dict_of_population[f"gen_{generation}"]["params"].append(params)
        else:
            self.dict_of_population[f"gen_{generation}"] = {
                "fitness": [[fitness]],
                "params": [[params]],
            }

    def get_metric_df(self):
        if self.metric_df is not None:
            logger("Metric df already exists")
        else:
            population_max = []
            for i in range(self.num_generations + 1):
                gen_value = np.max(self.dict_of_population[f"gen_{i}"]["fitness"])
                if i == 0:
                    population_max = [gen_value]
                else:
                    population_max.append(max(population_max[i - 1], gen_value))
            self.metric_df = pd.DataFrame(
                list(zip([i for i in range(self.num_generations + 1)], population_max)),
                columns=[GENERATION_COL, FITNESS_COL],
            )
        if self.mutation_df is not None:
            logger.info("Mutation df already exists")
        else:
            dfs = []
            for gen in self.mutation_changes:
                cur_df_dict = {
                    PARAMS_DIST_COL: self.mutation_changes[gen]["params_dist"],
                    FITNESS_DIFF_COL: self.mutation_changes[gen]["fitness_diff"],
                    "original_params": self.mutation_changes[gen]["original_params"],
                    "mutated_params": self.mutation_changes[gen]["mutated_params"],
                }
                cur_df = pd.DataFrame(cur_df_dict)
                cur_df[GENERATION_COL] = gen
                dfs.append(cur_df)
            if len(dfs) > 0:
                self.mutation_df = pd.concat(dfs)
            else:
                # warnings.warn("No mutations changes have been found to save", RuntimeWarning)
                self.mutation_df = pd.DataFrame([])
        if self.crossover_df is not None:
            logger.info("Crossover df already exists")
        else:
            dfs = []
            for gen in self.crossover_changes:
                cur_df_dict = {
                    "parent_1_params": self.crossover_changes[gen]["parent_1_params"],
                    "parent_2_params": self.crossover_changes[gen]["parent_2_params"],
                    "child_1_params": self.crossover_changes[gen]["child_1_params"],
                    "parent_1_fitness": self.crossover_changes[gen]["parent_1_fitness"],
                    "parent_2_fitness": self.crossover_changes[gen]["parent_2_fitness"],
                    "child_1_fitness": self.crossover_changes[gen]["child_1_fitness"],
                }
                # TODO: save child 2 params
                cur_df = pd.DataFrame(cur_df_dict)
                cur_df[GENERATION_COL] = gen
                dfs.append(cur_df)
            if len(dfs) > 0:
                self.crossover_df = pd.concat(dfs)
            else:
                warnings.warn("No crossover changes have been found to save", RuntimeWarning)
                self.crossover_df = pd.DataFrame([])

    def write_metrics_to_file(self):
        os.makedirs(self.save_path, exist_ok=True)
        self.metric_df.to_csv(
            os.path.join(
                self.save_path, f"{self.save_fname}_metric_{int(time.time())}.csv"
            )
        )
        self.mutation_df.to_csv(
            os.path.join(
                self.save_path, f"{self.save_fname}_mutation_{int(time.time())}.csv"
            )
        )
        self.crossover_df.to_csv(
            os.path.join(
                self.save_path, f"{self.save_fname}_crossover_{int(time.time())}.csv"
            )
        )

    def save_trace(self):
        self.get_metric_df()
        self.write_metrics_to_file()

    def save_and_visualise_trace(self, plot_mutation_effectiveness=False):
        self.save_trace()

        # traces vis
        graph_template = "plotly_white"

        fig = px.line(
            self.metric_df,
            x=GENERATION_COL,
            y=FITNESS_COL,
            title="Score changes with generations",
            template=graph_template,
        )
        fig.show()

        if plot_mutation_effectiveness:
            fig = px.scatter(
                self.mutation_df,
                x=PARAMS_DIST_COL,
                y=FITNESS_DIFF_COL,
                title="Effectiveness of mutation operation",
                template=graph_template,
            )
            fig.show()



from autotm.algorithms_for_tuning.individuals import Individual


class StatisticsCollector:
    """
    This logger handles collection of statistics
    """

    def log_iteration(self, evaluations: int, best_fitness: float):
        """
        :param evaluations: the number of used evaluations
        :param best_fitness: the best fitness in the current iteration
        """
        pass

    def log_individual(self, individual: Individual):
        """
        :param individual: a new evaluated individual
        """
        pass



import logging
import os
from typing import List, Union

import artm
import math
import pickle
import pandas as pd
import re

from autotm.preprocessing import PREPOCESSED_DATASET_FILENAME, RESERVED_TUPLE
from autotm.preprocessing.cooc import calculate_cooc
from autotm.utils import parallelize_dataframe
import itertools
from collections import Counter

logger = logging.getLogger(__name__)


def get_words_dict(text, stop_list):
    all_words = text
    words = sorted(set(all_words) - stop_list)
    return {w: all_words.count(w) for w in words}


def vocab_preparation(VOCAB_PATH, DICTIONARY_PATH):
    if not os.path.exists(VOCAB_PATH):
        with open(DICTIONARY_PATH, "r") as dictionary_file:
            with open(VOCAB_PATH, "w") as vocab_file:
                dictionary_file.readline()
                dictionary_file.readline()
                for line in dictionary_file:
                    elems = re.split(", ", line)
                    vocab_file.write(" ".join(elems[:2]) + "\n")


def _calculate_token_count():
    raise NotImplementedError


def _add_word_to_dict(word, w_dict):
    if word in w_dict:
        w_dict[word] += 1
    else:
        w_dict[word] = 1
    return w_dict


def _calculate_cooc_df_dict(data: list, vocab: List[str], window: int = 10) -> dict:
    cooc_df_dict = {}  # format dict{(tuple): cooc}
    term_freq_dict = {}
    existing_vocab = set(vocab)
    for text in data:
        document_cooc_df_dict = {}
        splitted = [word for word in text.split() if word in existing_vocab]
        for i in range(0, len(splitted) - window):
            for comb in itertools.combinations(splitted[i: i + window], 2):  # speed up
                comb = tuple(sorted(comb))  # adding comb sorting
                if comb in document_cooc_df_dict:
                    continue
                else:
                    document_cooc_df_dict[comb] = 1
                    term_freq_dict = _add_word_to_dict(comb[0], term_freq_dict)
                    term_freq_dict = _add_word_to_dict(comb[1], term_freq_dict)
        cooc_df_dict = dict(Counter(document_cooc_df_dict) + Counter(cooc_df_dict))
    return cooc_df_dict, term_freq_dict


def _calculate_cooc_tf_dict(data: list, vocab: List[str], window: int = 10) -> dict:
    term_freq_dict = {}
    cooc_tf_dict = {RESERVED_TUPLE: 0}  # format dict{(tuple): cooc}
    existing_vocab = set(vocab)
    for text in data:
        document_cooc_tf_dict = {}
        splitted = [word for word in text.split() if word in existing_vocab]
        for i in range(0, len(splitted) - window):
            for comb in itertools.combinations(splitted[i: i + window], 2):
                if comb in document_cooc_tf_dict:
                    document_cooc_tf_dict[comb] += 1
                else:
                    document_cooc_tf_dict[comb] = 1

                term_freq_dict = _add_word_to_dict(comb[0], term_freq_dict)
                term_freq_dict = _add_word_to_dict(comb[1], term_freq_dict)
                cooc_tf_dict[RESERVED_TUPLE] += 2
        # cooc_tf_dict = dict(Counter(document_cooc_tf_dict) + Counter(cooc_tf_dict))
        counter = Counter(document_cooc_tf_dict)
        counter.update(cooc_tf_dict)
        cooc_tf_dict = dict(counter)
    return cooc_tf_dict, term_freq_dict


def read_vocab(vocab_path: str) -> List[str]:
    # TODO: rewrite this part in case of several modalities
    vocab_words = []
    with open(vocab_path) as vpath:
        for line in vpath:
            splitted_line = line.split()
            if len(splitted_line) > 2:
                raise Exception("There are more than 2 modalities!")
            vocab_words.append(splitted_line[0].strip())

    return vocab_words


def calculate_ppmi(cooc_dict_path, n, term_freq_dict):
    logger.info("Calculating pPMI...")
    ppmi_dict = {}
    with open(cooc_dict_path) as fopen:
        for line in fopen:
            splitted_line = line.split()
            ppmi_dict[splitted_line[0]] = [
                f'{word.split(":")[0].strip()}:{max(math.log2((float(word.split(":")[1]) / n) / (term_freq_dict[word.split(":")[0].strip()] / n * term_freq_dict[splitted_line[0]] / n)), 0)}'
                for word in splitted_line[1:]
            ]
    return ppmi_dict


# TODO: rewrite to storing in rb tree
def calculate_cooc_dicts(vocab: List[str], df: pd.DataFrame, window=10, n_cores=-1):
    """

    :param vocab: list of words to be used for calculating cooccurrences
    :param df: dataframe with 'processed_text'  column
    :param window: The size of window to collect cooccurrences in
    :param n_cores: available cores for parallel processing. Default: -1 (all)
    :return: cooc_df and cooc_tf dictionaries
    """
    data = df["processed_text"].tolist()
    cooc_df_dict = parallelize_dataframe(
        data,
        _calculate_cooc_df_dict,
        n_cores,
        return_type="dict",
        vocab=vocab,
        window=window,
    )
    cooc_tf_dict = parallelize_dataframe(
        data,
        _calculate_cooc_tf_dict,
        n_cores,
        return_type="dict",
        vocab=vocab,
        window=window,
    )
    return cooc_df_dict, cooc_tf_dict


def write_vw_dict(res_dict, vocab_words, fpath):
    with open(fpath, "w") as fopen:
        for word in vocab_words:
            try:
                fopen.write(f"{word}" + " " + " ".join(res_dict[word]) + "\n")
            except:
                pass
    logger.info(f"{fpath} is ready!")


def convert_to_vw_format_and_save(cooc_dict, vocab_words, vw_path):
    if isinstance(cooc_dict, tuple):
        t_cooc_dict = cooc_dict[0]
    else:
        t_cooc_dict = cooc_dict
    data_dict = {}
    for item in sorted(t_cooc_dict.items(), key=lambda key: key[0]):
        if item == RESERVED_TUPLE:
            continue
        # word_1 = item[0][0]  # TODO: check this
        # word_2 = item[0][1]
        # if vocab_words.index(item[0][0]) > vocab_words.index(item[0][1]):
        #     word_2 = item[0][0]
        #     word_1 = item[0][1]
        if item[0][0] in data_dict:
            data_dict[item[0][0]].append(f"{item[0][1]}:{item[1]}")
        else:
            data_dict[item[0][0]] = [f"{item[0][1]}:{item[1]}"]
    write_vw_dict(data_dict, vocab_words, vw_path)


def prepearing_cooc_dict(
    BATCHES_DIR,
    WV_PATH,
    VOCAB_PATH,
    COOC_DICTIONARY_PATH,
    path_to_dataset,
    cooc_file_path_tf,
    cooc_file_path_df,
    ppmi_dict_tf,
    ppmi_dict_df,
    cooc_min_tf=0,
    cooc_min_df=0,
    cooc_window=10,
    n_cores=-1,
):
    """
    :param WV_PATH: path where to store data in Vowpal Wabbit format (https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Input-format)
    :param VOCAB_PATH: path where the full dictionary is listed
    :param COOC_DICTIONARY_PATH: path to cooccurrence dictionary
    :param path_to_dataset: path to folder with dataset
    :param cooc_file_path_tf: path to tf coocurrances
    :param cooc_file_path_df: path to df coocurrances
    :param ppmi_dict_tf: path to ppmi tf dict
    :param ppmi_dict_df: path to ppmi df dict
    :param cooc_min_tf: Minimal number of documents to cooc in pairs to store the results
    :param cooc_min_df: Minimal number of documents to docs to cooc to store the results
    :param cooc_window: size of the window where to search for the cooccurrences
    :return:
    """

    vocab_words = read_vocab(VOCAB_PATH)

    data = pd.read_csv(path_to_dataset)
    docs_count = data.shape[0]

    logger.debug("Performing calculate_cooc_dicts")
    cooc_dicts = calculate_cooc(batches_path=BATCHES_DIR, vocab=vocab_words, window_size=cooc_window)
    logger.debug("Performed calculate_cooc_dicts")
    cooc_df_dict, cooc_df_term_dict = cooc_dicts.cooc_df, cooc_dicts.cooc_df_term
    cooc_tf_dict, cooc_tf_term_dict = cooc_dicts.cooc_tf, cooc_dicts.cooc_tf_term

    pairs_count = cooc_tf_dict[RESERVED_TUPLE]

    del cooc_tf_dict[RESERVED_TUPLE]

    convert_to_vw_format_and_save(cooc_df_dict, vocab_words, cooc_file_path_df)
    convert_to_vw_format_and_save(cooc_tf_dict, vocab_words, cooc_file_path_tf)

    logger.debug("Performing calculate_ppmi")
    ppmi_df = calculate_ppmi(cooc_file_path_df, docs_count, cooc_df_term_dict)
    ppmi_tf = calculate_ppmi(cooc_file_path_tf, pairs_count, cooc_tf_term_dict)
    logger.debug("Performed calculate_ppmi")

    write_vw_dict(ppmi_tf, vocab_words, ppmi_dict_tf)
    write_vw_dict(ppmi_df, vocab_words, ppmi_dict_df)

    # ! bigartm - c $WV_PATH - v $VOCAB_PATH - -cooc - window
    # 10 - -write - cooc - tf $cooc_file_path_tf - -write - cooc - df $cooc_file_path_df - -write - ppmi - tf $ppmi_dict_tf - -write - ppmi - df $ppmi_dict_df

    cooc_dict = artm.Dictionary()
    cooc_dict.gather(
        data_path=BATCHES_DIR,
        cooc_file_path=ppmi_dict_tf,
        vocab_file_path=VOCAB_PATH,
        symmetric_cooc_values=True,
    )
    cooc_dict.save_text(COOC_DICTIONARY_PATH)


def return_string_part(name_type, text):
    tokens = text.split()
    tokens = [item for item in tokens if item != ""]

    return " |" + name_type + " " + " ".join(["{}:1".format(token) for token in tokens])


def prepare_voc(batches_dir, vw_path, dataset: Union[pd.DataFrame, str], column_name="processed_text.txt"):
    with open(vw_path, "w", encoding="utf8") as ofile:
        if isinstance(dataset, str):
            num_parts = 0
            try:
                for file in os.listdir(dataset):
                    if file.startswith("part"):
                        logger.debug("Dataset part_{}".format(num_parts))
                        if file.split(".")[-1] == "csv":
                            part = pd.read_csv(os.path.join(dataset, file))
                        else:
                            part = pd.read_parquet(os.path.join(dataset, file))
                        part_processed = part[column_name].tolist()
                        for text in part_processed:
                            result = return_string_part("@default_class", text)
                            ofile.write(result + "\n")
                        num_parts += 1

            except NotADirectoryError:
                logger.debug("Dataset part 1/1")
                part = pd.read_csv(dataset)
                part_processed = part[column_name].tolist()
                for text in part_processed:
                    result = return_string_part("@default_class", text)
                    ofile.write(result + "\n")
        else:
            part_processed = dataset[column_name].tolist()
            for text in part_processed:
                result = return_string_part("@default_class", text)
                ofile.write(result + "\n")

    logger.info(" batches {} \n vocabulary {} \n are ready".format(batches_dir, vw_path))


def prepare_batch_vectorizer(
    batches_dir: str, vw_path: str, dataset: Union[pd.DataFrame, str], column_name: str = "processed_text"
):
    prepare_voc(batches_dir, vw_path, dataset, column_name=column_name)
    batch_vectorizer = artm.BatchVectorizer(
        data_path=vw_path,
        data_format="vowpal_wabbit",
        target_folder=batches_dir,
        batch_size=100,
    )

    return batch_vectorizer


def mutual_info_dict_preparation(fname):
    tokens_dict = {}

    with open(fname) as handle:
        for ix, line in enumerate(handle):
            list_of_words = line.strip().split()
            word_1 = list_of_words[0]
            for word_val in list_of_words[1:]:
                word_2, value = word_val.split(":")
                tokens_dict["{}_{}".format(word_1, word_2)] = float(value)
                tokens_dict["{}_{}".format(word_2, word_1)] = float(value)
    return tokens_dict


def prepare_all_artifacts(save_path: str):
    DATASET_PATH = os.path.join(save_path, PREPOCESSED_DATASET_FILENAME)
    BATCHES_DIR = os.path.join(save_path, "batches")
    WV_PATH = os.path.join(save_path, "test_set_data_voc.txt")
    COOC_DICTIONARY_PATH = os.path.join(save_path, "cooc_dictionary.txt")
    DICTIONARY_PATH = os.path.join(save_path, "dictionary.txt")
    VOCAB_PATH = os.path.join(save_path, "vocab.txt")
    cooc_file_path_df = os.path.join(save_path, "cooc_df.txt")
    cooc_file_path_tf = os.path.join(save_path, "cooc_tf.txt")
    ppmi_dict_df = os.path.join(save_path, "ppmi_df.txt")
    ppmi_dict_tf = os.path.join(save_path, "ppmi_tf.txt")
    MUTUAL_INFO_DICT_PATH = os.path.join(save_path, "mutual_info_dict.pkl")
    DOCUMENTS_TO_BATCH_PATH = os.path.join(save_path, PREPOCESSED_DATASET_FILENAME)

    # TODO: check why batch vectorizer is returned (unused further)
    logger.debug("Starting batch vectorizer...")
    prepare_batch_vectorizer(
        BATCHES_DIR, WV_PATH, DOCUMENTS_TO_BATCH_PATH
    )

    logger.debug("Preparing artm.Dictionary...")
    my_dictionary = artm.Dictionary()
    my_dictionary.gather(data_path=BATCHES_DIR, vocab_file_path=WV_PATH)
    my_dictionary.filter(min_df=3, class_id="text")
    my_dictionary.save_text(DICTIONARY_PATH)

    logger.debug("Vocabulary preparing...")
    vocab_preparation(VOCAB_PATH, DICTIONARY_PATH)

    logger.debug("Cooc dictionary preparing...")
    prepearing_cooc_dict(
        BATCHES_DIR,
        WV_PATH,
        VOCAB_PATH,
        COOC_DICTIONARY_PATH,
        DATASET_PATH,
        cooc_file_path_tf,
        cooc_file_path_df,
        ppmi_dict_tf,
        ppmi_dict_df,
    )

    logger.debug("Mutual info dictionary preparing...")
    mutual_info_dict = mutual_info_dict_preparation(ppmi_dict_tf)
    with open(MUTUAL_INFO_DICT_PATH, "wb") as handle:
        pickle.dump(mutual_info_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)



#!/usr/bin/env python3
import logging
import logging.config
import sys
import uuid
from typing import Union, Optional

from autotm.algorithms_for_tuning.genetic_algorithm.statistics_collector import StatisticsCollector
from autotm.algorithms_for_tuning.genetic_algorithm.ga import GA
from autotm.algorithms_for_tuning.genetic_algorithm.surrogate import Surrogate
from autotm.algorithms_for_tuning.individuals import IndividualBuilder
from autotm.fitness.estimator import ComputableFitnessEstimator, SurrogateEnabledComputableFitnessEstimator, \
    DistributedSurrogateEnabledComputableFitnessEstimator
from autotm.fitness.tm import fit_tm, TopicModel
from autotm.utils import make_log_config_dict

logger = logging.getLogger(__name__)

NUM_FITNESS_EVALUATIONS = 150
SPEEDUP = True


def get_best_individual(
        dataset: str,
        data_path: str,
        exp_id: Union[int, str],
        topic_count: int,
        num_individuals: int = 11,
        num_iterations: int = 400,
        num_fitness_evaluations: int = None,
        mutation_type: str = "psm",
        crossover_type: str = "blend_crossover",
        selection_type: str = "fitness_prop",
        elem_cross_prob: float = None,
        cross_alpha: float = 0.5046,
        best_proc: float = 0.4439,
        log_file: str = "/var/log/tm-alg.log",
        tag: str = "v0",
        surrogate_name: str = None,  # fix
        gpr_kernel: str = None,
        gpr_alpha: float = None,
        gpr_normalize_y: float = None,
        use_pipeline: bool = False,
        use_nelder_mead_in_mutation: bool = False,
        use_nelder_mead_in_crossover: bool = False,
        use_nelder_mead_in_selector: bool = False,
        train_option: str = "offline",
        quiet_log: bool = False,
        statistics_collector: Optional[StatisticsCollector] = None,
        individual_type: str = "regular",
        fitness_estimator_type: str = "local", # distributed
        **kwargs
):
    """

    :param dataset: Dataset name that is being processed. The name will be used to store results
    :param data_path: Path to all the artifacts obtained after
    :param exp_id: Mlflow experiment id
    :param topic_count: desired count of SPECIFIC topics (in optimization process BACK topics are also produced, thus the total amount of topics can be more than topic_count)
    :param num_individuals: Number of individuals in generation
    :param num_iterations: Number of iterations to make
    :param num_fitness_evaluations: Max number of possible fitness estimations. This setting may lead to premature algorithm stopping even if there is more generations to go
    :param mutation_type: Mutation type can have value from (mutation_one_param, combined, psm, positioning_mutation)
    :param crossover_type: Crossover type can have value from (crossover_pmx, crossover_one_point, blend_crossover)
    :param selection_type: Selection type can have value from (fitness_prop, rank_based)
    :param elem_cross_prob: Сrossover probability for each of the elements
    :param cross_alpha: Alpha parameter for blend crossover
    :param best_proc: Number of best parents to propagate
    :param log_file: A log file to write logs of the algorithm execution to
    :param tag: Service tag for massive experiments to store the models with tags
    :param surrogate_name: Name of the surrogate model
    :param gpr_kernel: Kernel name for gpr surrogate
    :param gpr_alpha: Alpha parameter for gpr
    :param gpr_normalize_y: y normalization parameter for gpr
    :param use_nelder_mead_in_mutation:
    :param use_nelder_mead_in_crossover:
    :param use_nelder_mead_in_selector:
    :return:
    """

    assert (
            sum(
                [
                    use_nelder_mead_in_mutation,
                    use_nelder_mead_in_crossover,
                    use_nelder_mead_in_selector,
                ]
            )
            <= 1
    )

    logger.debug(f"Command line: {sys.argv}")

    run_uid = str(uuid.uuid4())
    tag = tag if tag is not None else str(run_uid)
    logging_config = make_log_config_dict(filename=log_file, uid=run_uid, quiet=quiet_log)
    logging.config.dictConfig(logging_config)

    logger.info(f"Starting a new run of algorithm. Args: {sys.argv[1:]}")

    if elem_cross_prob is not None:
        elem_cross_prob = float(elem_cross_prob)

    if cross_alpha is not None:
        cross_alpha = float(cross_alpha)

    ibuilder = IndividualBuilder(individual_type)

    if fitness_estimator_type == "local" and surrogate_name:
        fitness_estimator = SurrogateEnabledComputableFitnessEstimator(
            ibuilder,
            Surrogate(surrogate_name),
            "type1",
            SPEEDUP,
            num_fitness_evaluations,
            statistics_collector
        )
    elif fitness_estimator_type == "local":
        fitness_estimator = ComputableFitnessEstimator(ibuilder, num_fitness_evaluations, statistics_collector)
    elif fitness_estimator_type == "distributed" and surrogate_name:
        fitness_estimator = DistributedSurrogateEnabledComputableFitnessEstimator(
            ibuilder,
            Surrogate(surrogate_name),
            "type1",
            SPEEDUP,
            num_fitness_evaluations,
            statistics_collector
        )
    elif fitness_estimator_type == "distributed":
        fitness_estimator = ComputableFitnessEstimator(ibuilder, num_fitness_evaluations, statistics_collector)
    else:
        raise ValueError("Incorrect settings")

    g = GA(
        dataset=dataset,
        data_path=data_path,
        num_individuals=num_individuals,
        num_iterations=num_iterations,
        ibuilder=ibuilder,
        fitness_estimator=fitness_estimator,
        mutation_type=mutation_type,
        crossover_type=crossover_type,
        selection_type=selection_type,
        elem_cross_prob=elem_cross_prob,
        num_fitness_evaluations=num_fitness_evaluations,
        best_proc=best_proc,
        alpha=cross_alpha,
        exp_id=exp_id,
        topic_count=topic_count,
        tag=tag,
        surrogate_name=surrogate_name,
        gpr_kernel=gpr_kernel,
        gpr_alpha=gpr_alpha,
        normalize_y=gpr_normalize_y,
        use_pipeline=use_pipeline,
        use_nelder_mead_in_mutation=use_nelder_mead_in_mutation,
        use_nelder_mead_in_crossover=use_nelder_mead_in_crossover,
        use_nelder_mead_in_selector=use_nelder_mead_in_selector,
        train_option=train_option,
        statistics_collector=statistics_collector,
        **kwargs
    )
    best_individual = g.run(verbose=True)
    logger.info(f"Best individual fitness_value: {best_individual.fitness_value * (-1)}")

    return best_individual


def run_algorithm(
        dataset: str,
        data_path: str,
        exp_id: Union[int, str],
        topic_count: int,
        num_individuals: int = 11,
        num_iterations: int = 50,
        num_fitness_evaluations: int = None,
        mutation_type: str = "psm",
        crossover_type: str = "blend_crossover",
        selection_type: str = "fitness_prop",
        elem_cross_prob: float = None,
        cross_alpha: float = 0.5046,
        best_proc: float = 0.4439,
        log_file: str = "/var/log/tm-alg.log",
        tag: str = "v0",
        surrogate_name: str = None,  # fix
        gpr_kernel: str = None,
        gpr_alpha: float = None,
        gpr_normalize_y: float = None,
        use_pipeline: bool = True,
        use_nelder_mead_in_mutation: bool = False,
        use_nelder_mead_in_crossover: bool = False,
        use_nelder_mead_in_selector: bool = False,
        train_option: str = "offline",
        quiet_log: bool = False,
        individual_type: str = "regular",
        fitness_estimator_type: str = "local"
) -> TopicModel:
    best_individual = get_best_individual(dataset, data_path, exp_id, topic_count, num_individuals, num_iterations,
                                          num_fitness_evaluations, mutation_type, crossover_type, selection_type,
                                          elem_cross_prob, cross_alpha, best_proc, log_file, tag, surrogate_name,
                                          gpr_kernel, gpr_alpha, gpr_normalize_y, use_pipeline,
                                          use_nelder_mead_in_mutation, use_nelder_mead_in_crossover,
                                          use_nelder_mead_in_selector, train_option, quiet_log,
                                          individual_type=individual_type,
                                          fitness_estimator_type=fitness_estimator_type)

    best_topic_model = fit_tm(
        preproc_data_path=data_path,
        topic_count=topic_count,
        params=best_individual.params,
        train_option=train_option
    )

    return best_topic_model



from abc import ABC, abstractmethod

from typing import List


class AbstractParams(ABC):
    @property
    @abstractmethod
    def basic_topics(self) -> int:
        ...

    @property
    @abstractmethod
    def mutation_probability(self):
        ...

    @abstractmethod
    def make_params_dict(self):
        ...

    @abstractmethod
    def run_train(self, model):
        """
        Trains the topic model
        :param model: an instance of TopicModel
        :return:
        """
        ...

    @abstractmethod
    def validate_params(self) -> bool:
        ...

    @abstractmethod
    def crossover(self, parent2: "AbstractParams", **kwargs) -> List["AbstractParams"]:
        ...

    @abstractmethod
    def mutate(self, **kwargs) -> "AbstractParams":
        ...

    @abstractmethod
    def to_vector(self) -> List[float]:
        ...




import random
from typing import List, Union

from pydantic import BaseModel


class IntRangeDistribution(BaseModel):
    low: int
    high: int

    def create_value(self) -> int:
        return random.randint(self.low, self.high)

    def clip(self, value) -> int:
        value = int(value)
        return int(max(self.low, min(self.high, value)))


class FloatRangeDistribution(BaseModel):
    low: float
    high: float

    def create_value(self) -> float:
        return random.uniform(self.low, self.high)

    def clip(self, value) -> float:
        return max(self.low, min(self.high, value))


class Param(BaseModel):
    """
    Single parameter of a stage.
    Distribution can be unavailable after serialisation.
    """
    name: str
    distribution: Union[IntRangeDistribution, FloatRangeDistribution]

    def create_value(self):
        if self.distribution is None:
            raise ValueError("Distribution is unavailable. One must restore the distribution after serialisation.")
        return self.distribution.create_value()


class StageType(BaseModel):
    """
    Stage template that defines params for a stage.
    See create_stage generator.
    """
    name: str
    params: List[Param]


class Stage(BaseModel):
    """
    Stage instance with parameter values.
    """
    stage_type: StageType
    values: List

    def __init__(self, **data):
        """
        :param stage_type: back reference to the template for parameter mutation
        :param values: instance's parameter values
        """
        super().__init__(**data)
        if len(self.stage_type.params) != len(self.values):
            raise ValueError("Number of values does not match number of parameters.")

    def __str__(self):
        return f"{self.stage_type.name}{self.values}"

    def clip_values(self):
        self.values = [param.distribution.clip(value) for param, value in zip(self.stage_type.params, self.values)]


def create_stage(stage_type: StageType) -> Stage:
    return Stage(stage_type=stage_type, values=[param.create_value() for param in stage_type.params])


class Pipeline(BaseModel):
    """
    List of stages that can be mutated.
    """
    stages: List[Stage]
    required_params: Stage

    def __str__(self):
        return f'{str(self.required_params)} {" ".join(map(str, self.stages))}'

    def random_stage_index(self, with_last: bool = False):
        last = len(self.stages)
        if with_last:
            last += 1
        return random.randint(0, last - 1)

    def __lt__(self, other):
        # important for sort method usage
        return len(self.stages) < len(other.stages)



import logging
import os
import pprint
import tempfile
from contextlib import contextmanager
from typing import Optional

import click
import pandas as pd
import yaml

from autotm.base import AutoTM

# TODO: add proper logging format initialization
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()


def obtain_autotm_params(
        config_path: str,
        topic_count: Optional[int],
        lang: Optional[str],
        alg: Optional[str],
        surrogate_alg: Optional[str],
        log_file: Optional[str]
):
    # TODO: define clearly format of config
    if config_path is not None:
        logger.info(f"Reading config from path: {os.path.abspath(config_path)}")
        with open(config_path, "r") as f:
            config = yaml.safe_load(f)
    else:
        config = dict()

    if topic_count is not None:
        config['topic_count'] = topic_count
    if lang is not None:
        pp = config.get('preprocessing_params', dict())
        pp['lang'] = lang
        config['preprocessing_params'] = pp
    if alg is not None:
        config['alg_name'] = alg
    if surrogate_alg is not None:
        config['surrogate_alg_name'] = surrogate_alg
    if log_file is not None:
        config['log_file_path'] = log_file

    return config


@contextmanager
def prepare_working_dir(working_dir: Optional[str] = None):
    if working_dir is None:
        with tempfile.TemporaryDirectory(prefix="autotm_wd_") as tmp_working_dir:
            yield tmp_working_dir
    else:
        yield working_dir


@click.group()
@click.option('-v', '--verbose', is_flag=True, show_default=True, default=False, help="Verbose output")
def cli(verbose: bool):
    if verbose:
        logging.basicConfig(level=logging.DEBUG, force=True)


@cli.command()
@click.option('--config', 'config_path', type=str, help="A path to config for fitting the model")
@click.option(
    '--working-dir',
    type=str,
    help="A path to working directory used by AutoTM for storing intermediate files. "
         "If not specified temporary directory will be created in the current directory "
         "and will be deleted upon successful finishing."
)
@click.option('--in', 'in_', type=str, required=True, help="A file in csv format with text corpus to build model on")
@click.option(
    '--out',
    type=str,
    default="mixtures.csv",
    help="A path to a file in csv format that will contain topic mixtures for texts from the incoming corpus"
)
@click.option('--model', type=str, default='model.artm', help="A path that will contain fitted ARTM model")
@click.option('-t', '--topic-count', type=int, help="Number of topics to fit model with")
@click.option('--lang', type=str, help='Language of the dataset')
@click.option('--alg', type=str, help="Hyperparameters tuning algorithm. Available: ga, bayes")
@click.option('--surrogate-alg', type=str, help="Surrogate algorithm to use.")
@click.option('--log-file', type=str, help="Log file path")
@click.option(
    '--overwrite',
    is_flag=True,
    show_default=True,
    default=False,
    help="Overwrite if model or/and mixture files already exist"
)
def fit(
        config_path: Optional[str],
        working_dir: Optional[str],
        in_: str,
        out: str,
        model: str,
        topic_count: Optional[int],
        lang: Optional[str],
        alg: Optional[str],
        surrogate_alg: Optional[str],
        log_file: Optional[str],
        overwrite: bool
):
    config = obtain_autotm_params(config_path, topic_count, lang, alg, surrogate_alg, log_file)

    logger.debug(f"Running AutoTM with params: {pprint.pformat(config, indent=4)}")

    logger.info(f"Loading data from {os.path.abspath(in_)}")
    df = pd.read_csv(in_)

    with prepare_working_dir(working_dir) as work_dir:
        logger.info(f"Using working directory {os.path.abspath(work_dir)} for AutoTM")
        autotm = AutoTM(
            **config,
            working_dir_path=work_dir
        )
        mixtures = autotm.fit_predict(df)

        logger.info(f"Saving model to {os.path.abspath(model)}")
        autotm.save(model, overwrite=overwrite)
        logger.info(f"Saving mixtures to {os.path.abspath(out)}")
        mixtures.to_csv(out, mode='w' if overwrite else 'x')

    logger.info("Finished AutoTM")


@cli.command()
@click.option('--model', type=str, required=True, help="A path to fitted saved ARTM model")
@click.option('--in', 'in_', type=str, required=True, help="A file in csv format with text corpus to build model on")
@click.option(
    '--out',
    type=str,
    default="mixtures.csv",
    help="A path to a file in csv format that will contain topic mixtures for texts from the incoming corpus"
)
@click.option(
    '--overwrite',
    is_flag=True,
    show_default=True,
    default=False,
    help="Overwrite if the mixture file already exists"
)
def predict(model: str, in_: str, out: str, overwrite: bool):
    logger.info(f"Loading model from {os.path.abspath(model)}")
    autotm_loaded = AutoTM.load(model)

    logger.info(f"Predicting mixtures for data from {os.path.abspath(in_)}")
    df = pd.read_csv(in_)
    mixtures = autotm_loaded.predict(df)

    logger.info(f"Saving mixtures to {os.path.abspath(out)}")
    mixtures.to_csv(out, mode='w' if overwrite else 'x')


if __name__ == "__main__":
    cli()



import logging
from typing import List, Optional

from autotm.algorithms_for_tuning.individuals import Individual, IndividualBuilder
from autotm.fitness.tm import fit_tm_of_individual
from autotm.params_logging_utils import log_params_and_artifacts, log_stats, model_files
from autotm.schemas import IndividualDTO

logger = logging.getLogger("root")


def do_fitness_calculating(
        individual: IndividualDTO,
        log_artifact_and_parameters: bool = False,
        log_run_stats: bool = False,
        alg_args: Optional[str] = None,
        is_tmp: bool = False,
) -> IndividualDTO:
    # make copy
    individual_json = individual.model_dump_json()
    individual = IndividualDTO.model_validate_json(individual_json)
    logger.info("Doing fitness calculating with individual %s" % individual_json)

    with fit_tm_of_individual(
            dataset=individual.dataset,
            data_path=individual.data_path,
            params=individual.params,
            fitness_name=individual.fitness_name,
            topic_count=individual.topic_count,
            force_dataset_settings_checkout=individual.force_dataset_settings_checkout,
            train_option=individual.train_option,
    ) as (time_metrics, metrics, tm):
        individual.fitness_value = metrics

        with model_files(tm) as tm_files:
            if log_artifact_and_parameters:
                log_params_and_artifacts(
                    tm, tm_files, individual, time_metrics, alg_args, is_tmp=is_tmp
                )

            if log_run_stats:
                log_stats(tm, tm_files, individual, time_metrics, alg_args)

    return individual


def calculate_fitness(
        individual: IndividualDTO,
        log_artifact_and_parameters: bool = False,
        log_run_stats: bool = False,
        alg_args: Optional[str] = None,
        is_tmp: bool = False,
) -> IndividualDTO:
    return do_fitness_calculating(
        individual, log_artifact_and_parameters, log_run_stats, alg_args, is_tmp
    )


def estimate_fitness(ibuilder: IndividualBuilder, population: List[Individual]) -> List[Individual]:
    logger.info("Calculating fitness...")
    population_with_fitness = []
    for individual in population:
        if individual.dto.fitness_value is not None:
            logger.info("Fitness value already calculated")
            population_with_fitness.append(individual)
            continue
        individ_with_fitness = calculate_fitness(individual.dto)
        population_with_fitness.append(ibuilder.make_individual(individ_with_fitness))
    logger.info("The fitness results have been obtained")
    return population_with_fitness


def log_best_solution(
        ibuilder: IndividualBuilder,
        individual: Individual,
        wait_for_result_timeout: Optional[float] = None,
        alg_args: Optional[str] = None,
        is_tmp: bool = False,
):
    logger.info("Sending a best individual to be logged")
    res = ibuilder.make_individual(calculate_fitness(individual.dto,
                                            log_artifact_and_parameters=True,
                                            is_tmp=is_tmp))

    # TODO: write logging
    return res



from __future__ import print_function

import glob
import logging
import os
import sys
import time
from dataclasses import dataclass
from typing import Dict, Tuple, List, Set

import artm
from six.moves import range

from autotm.preprocessing import RESERVED_TUPLE

logger = logging.getLogger(__name__)


@dataclass
class CoocDictionaries:
    cooc_df: Dict[Tuple[str, str], float]
    cooc_tf: Dict[Tuple[str, str], float]
    cooc_df_term: Dict[str, float]
    cooc_tf_term: Dict[str, float]


def __create_batch_dictionary(batch):
    batch_dictionary = {}
    for index, token in enumerate(batch.token):
        batch_dictionary[index] = token

    return batch_dictionary


def __process_batch(
        global_cooc_df_dictionary,
        global_cooc_tf_dictionary,
        global_cooc_df_term_dictionary,
        global_cooc_tf_term_dictionary,
        batch,
        window_size,
        vocab: Set[str]):

    global_cooc_tf_dictionary[RESERVED_TUPLE] = 0.0
    batch_dictionary = __create_batch_dictionary(batch)

    def __process_window_df(global_cooc_dict, global_word_dict, token_ids,
                            doc_seen_pairs: Set[Tuple[int, int]], doc_seen_words: Set[int]):
        for j in range(1, len(token_ids)):
            token_1 = batch_dictionary[token_ids[0]]
            token_2 = batch_dictionary[token_ids[j]]

            if token_1 == token_2:
                continue

            if token_1 not in vocab or token_2 not in vocab:
                continue

            token_pair = (min(token_1, token_2), max(token_1, token_2))

            if token_pair not in doc_seen_pairs:
                global_cooc_dict[token_pair] = global_cooc_dict.get(token_pair, 0.0) + 1.0
                doc_seen_pairs.add(token_pair)

            if token_1 not in doc_seen_words:
                global_word_dict[token_1] = global_word_dict.get(token_1, 0.0) + 1.0
                doc_seen_words.add(token_1)

            if token_2 not in doc_seen_words:
                global_word_dict[token_2] = global_word_dict.get(token_2, 0.0) + 1.0
                doc_seen_words.add(token_2)

    def __process_window_tf(global_cooc_dict, global_word_dict, token_ids, token_weights: List[float]):
        for j in range(1, len(token_ids)):
            value = min(token_weights[0], token_weights[j])

            token_1 = batch_dictionary[token_ids[0]]
            token_2 = batch_dictionary[token_ids[j]]

            if token_1 == token_2:
                continue

            if token_1 not in vocab or token_2 not in vocab:
                continue

            token_pair = (min(token_1, token_2), max(token_1, token_2))
            global_cooc_dict[token_pair] = global_cooc_dict.get(token_pair, 0.0) + value
            global_word_dict[token_1] = global_word_dict.get(token_1, 0.0) + 1.0
            global_word_dict[token_2] = global_word_dict.get(token_2, 0.0) + 1.0
            global_cooc_dict[RESERVED_TUPLE] += 2

    for item in batch.item:
        doc_seen_pairs = set()
        doc_seen_words = set()
        real_window_size = window_size if window_size > 0 else len(item.token_id)
        for window_start_id in range(len(item.token_id)):
            end_index = window_start_id + real_window_size
            token_ids = item.token_id[window_start_id: end_index if end_index < len(item.token_id) else len(item.token_id)]
            token_weights = item.token_weight[window_start_id: end_index if end_index < len(item.token_id) else len(item.token_id)]
            __process_window_df(global_cooc_df_dictionary, global_cooc_df_term_dictionary, token_ids, doc_seen_pairs, doc_seen_words)
            __process_window_tf(global_cooc_tf_dictionary, global_cooc_tf_term_dictionary, token_ids, token_weights)


def __size(global_cooc_dictionary):
    result = sys.getsizeof(global_cooc_dictionary) \
             + sum(sys.getsizeof(k) + sys.getsizeof(v) for k, v in global_cooc_dictionary.items())

    return result


def calculate_cooc(batches_path: str, vocab: List[str], window_size: int=10) -> CoocDictionaries:
    global_time_start = time.time()
    batches_list = glob.glob(os.path.join(batches_path, '*.batch'))

    logger.info(
        "Calculating cooc: %s batches were found in %s, start processing"
        % (len(batches_list), batches_path)
    )

    global_cooc_df_dictionary = dict()
    global_cooc_tf_dictionary = dict()
    global_cooc_df_term_dictionary = dict()
    global_cooc_tf_term_dictionary = dict()
    for index, filename in enumerate(batches_list):
        local_time_start = time.time()
        logger.debug('Processing batch: %s' % index)
        current_batch = artm.messages.Batch()
        with open(filename, 'rb') as fin:
            current_batch.ParseFromString(fin.read())
        __process_batch(
            global_cooc_df_dictionary, global_cooc_tf_dictionary,
            global_cooc_df_term_dictionary, global_cooc_tf_term_dictionary,
            current_batch, window_size,
            set(vocab)
        )

        logger.debug('Finished batch, elapsed time: %s' % (time.time() - local_time_start))

    logger.info(
        'Finished cooc dict collection, elapsed time: %s, size: %s Gb'
        % (time.time() - global_time_start, __size(global_cooc_tf_dictionary) / 1000000000.0)
    )

    return CoocDictionaries(
        cooc_df=global_cooc_df_dictionary,
        cooc_tf=global_cooc_tf_dictionary,
        cooc_df_term=global_cooc_df_term_dictionary,
        cooc_tf_term=global_cooc_df_term_dictionary
    )



#!/usr/bin/env python3
import copy
import logging
import logging.config
import os
import random
import sys
import uuid
from multiprocessing.pool import AsyncResult
from typing import List, Optional, Union

import yaml
from hyperopt import STATUS_OK, fmin, hp, tpe
from tqdm import tqdm
from yaml import Loader

from autotm.algorithms_for_tuning.individuals import IndividualDTO, IndividualBuilder
from autotm.fitness.estimator import FitnessEstimator, ComputableFitnessEstimator
from autotm.fitness.tm import fit_tm, TopicModel
from autotm.params import FixedListParams
from autotm.utils import TqdmToLogger, make_log_config_dict

ALG_ID = "bo"

SPACE = {
    "decor": hp.quniform("decor", 0, 1e5, 0.05),
    "n1": hp.quniform("n1", 0, 8, 1),
    "spb": hp.quniform("spb", 1e-3, 1e2, 0.05),
    "stb": hp.quniform("stb", 1e-3, 1e2, 0.05),
    "n2": hp.quniform("n2", 0, 8, 1),
    "sp1": hp.quniform("sp1", -1e2, -1e-3, 0.05),
    "st1": hp.quniform("st1", -1e2, -1e-3, 0.05),
    "n3": hp.quniform("n3", 0, 8, 1),
    "sp2": hp.quniform("sp2", -1e2, -1e-3, 0.05),
    "st2": hp.quniform("st2", -1e2, -1e-3, 0.05),
    "n4": hp.quniform("n4", 0, 8, 1),
    "B": hp.quniform("B", 0, 8, 1),
    "decor_2": hp.quniform("decor_2", 0, 1e5, 0.05),
}

logger = logging.getLogger("BO")

# TODO: refactor this part
# getting config vars
if "FITNESS_CONFIG_PATH" in os.environ:
    filepath = os.environ["FITNESS_CONFIG_PATH"]
    with open(filepath, "r") as file:
        config = yaml.load(file, Loader=Loader)
else:
    config = {
        "testMode": False,
        "boAlgoParams": {
            "numEvals": 100
        }
    }


def estimate_fitness(
    population: List[IndividualDTO], _: bool = False, __: int = 2
) -> List[IndividualDTO]:
    results = []

    tqdm_out = TqdmToLogger(logger, level=logging.INFO)
    for p in tqdm(population, file=tqdm_out):
        individual = copy.deepcopy(p)
        individual.fitness_value = random.random()
        results.append(individual)

    return results


def log_best_solution(
    individual: IndividualDTO,
    wait_for_result_timeout: Optional[float] = None,
    alg_args: Optional[str] = None,
) -> Union[IndividualDTO, AsyncResult]:
    ind = copy.deepcopy(individual)
    ind.fitness_value = random.random()
    return ind


NUM_FITNESS_EVALUATIONS = config["boAlgoParams"]["numEvals"]


class BigartmFitness:
    def __init__(self,
                 data_path: str,
                 topic_count: int,
                 ibuilder: IndividualBuilder,
                 fitness_estimator: FitnessEstimator,
                 dataset: str,
                 exp_id: Optional[int] = None):
        self.data_path = data_path
        self.topic_count = topic_count
        self.ibuilder = ibuilder
        self.fitness_estimator = fitness_estimator
        self.dataset = dataset
        self.exp_id = exp_id

    def parse_kwargs(self, **kwargs):
        params = []
        params.append(kwargs.get("decor", 1))
        params.append(int(kwargs.get("n1", 1)))
        params.append(kwargs.get("spb", 1))
        params.append(kwargs.get("stb", 1))
        params.append(int(kwargs.get("n2", 1)))
        params.append(kwargs.get("sp1", 1))
        params.append(kwargs.get("st1", 1))
        params.append(int(kwargs.get("n3", 1)))
        params.append(kwargs.get("sp2", 1))
        params.append(kwargs.get("st2", 1))
        params.append(int(kwargs.get("n4", 1)))
        params.append(int(kwargs.get("B", 1)))
        params.append(kwargs.get("decor_2", 1))
        return params

    def make_ind_dto(self, **kwargs):
        # TODO: adapt this function to work with baesyian optimization
        params = [float(i) for i in self.parse_kwargs(**kwargs)]
        params = params[:-1] + [0.0, 0.0, 0.0] + [params[-1]]
        return IndividualDTO(
            id=str(uuid.uuid4()),
            data_path=self.data_path,
            dataset=self.dataset,
            topic_count=self.topic_count,
            params=FixedListParams(params=params),
            exp_id=self.exp_id,
            alg_id=ALG_ID,
        )

    def __call__(self, kwargs):
        population = [self.ibuilder.make_individual(self.make_ind_dto(**kwargs))]
        population = self.fitness_estimator.estimate(-1, population)
        individ = population[0]
        return {"loss": -1 * individ.fitness_value, "status": STATUS_OK}


def run_algorithm(dataset,
                  data_path,
                  topic_count,
                  log_file,
                  exp_id,
                  num_evaluations,
                  individual_type: str = "regular",
                  train_option: str = "offline") -> TopicModel:
    run_uid = uuid.uuid4() if not config["testMode"] else None
    logging_config = make_log_config_dict(filename=log_file, uid=run_uid)
    logging.config.dictConfig(logging_config)

    ibuilder = IndividualBuilder(individual_type)
    fitness_estimator = ComputableFitnessEstimator(ibuilder, num_evaluations)

    fitness = BigartmFitness(data_path, topic_count, ibuilder, fitness_estimator, dataset, exp_id)
    best_params = fmin(
        fitness, SPACE, algo=tpe.suggest, max_evals=num_evaluations
    )
    best_solution_dto = fitness.make_ind_dto(**best_params)
    best_solution_dto = log_best_solution(
        best_solution_dto, wait_for_result_timeout=-1, alg_args=" ".join(sys.argv)
    )

    best_topic_model = fit_tm(
        preproc_data_path=data_path,
        topic_count=topic_count,
        params=best_solution_dto.params,
        train_option=train_option
    )

    return best_topic_model



import logging
import os
import pickle
import shutil
import tempfile
import uuid
from typing import Union, Optional, Any, Dict, List

import artm
import pandas as pd
from sklearn.base import BaseEstimator

import warnings

from autotm.preprocessing import PREPOCESSED_DATASET_FILENAME

# TODO: Suppressing of DeprecationWarnings that are raise if we are running with __main__, need to research further
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from autotm.algorithms_for_tuning.bayesian_optimization import bayes_opt

from autotm.algorithms_for_tuning.genetic_algorithm import genetic_algorithm
from autotm.fitness.tm import extract_topics, print_topics
from autotm.infer import TopicsExtractor
from autotm.preprocessing.dictionaries_preparation import prepare_all_artifacts
from autotm.preprocessing.text_preprocessing import process_dataset, PROCESSED_TEXT_COLUMN

logger = logging.getLogger(__name__)


class AutoTM(BaseEstimator):
    _ARTM_MODEL_FILENAME = "artm_model"
    _AUTOTM_DATA_FILENAME = "autotm_data"
    _SUPPORTED_ALGS = ["ga", "bayes"]

    @classmethod
    def load(cls, path: str) -> 'AutoTM':
        """
        Loads AutoTM instance from a path on local filesystem.
        :param path: a local filesystem path to load an AutoTM instance from.
        """
        assert os.path.exists(path), f"Path doesn't exist: {path}"

        artm_model_path = os.path.join(path, cls._ARTM_MODEL_FILENAME)
        autotm_data_path = os.path.join(path, cls._AUTOTM_DATA_FILENAME)

        if not (os.path.exists(artm_model_path) and os.path.exists(autotm_data_path)):
            raise FileNotFoundError(f"One or two of the follwing paths don't exist: "
                                    f"{artm_model_path}, {autotm_data_path}")

        model = artm.load_artm_model(artm_model_path)

        with open(autotm_data_path, "rb") as f:
            params = pickle.load(f)

        autotm = AutoTM(**params)
        autotm._model = model

        return autotm

    def __init__(self,
                 topic_count: int = 10,
                 preprocessing_params: Optional[Dict[str, Any]] = None,
                 alg_name: str = "ga",
                 alg_params: Optional[Dict[str, Any]] = None,
                 surrogate_alg_name: Optional[str] = None,
                 surrogate_alg_params: Optional[Dict[str, Any]] = None,
                 artm_train_options: Optional[Dict[str, Any]] = None,
                 working_dir_path: Optional[str] = None,
                 texts_column_name: str = "text",
                 log_file_path: Optional[str] = None,
                 exp_id: Optional[str] = None,
                 exp_tag: Optional[str] = None,
                 exp_dataset_name: Optional[str] = None
                 ):
        """
        :param topic_count: Count of topics to fit ARTM model with
        :param preprocessing_params: A dict with params for the preprocessor
        :param alg_name: An algorithm to use for hyper parameters tuning of ARTM model (available: ga, bayes)
        :param alg_params: A dict with the algorithm specific parameters. Depends on alg_name.
            If not specified default parameters will be used.
        :param surrogate_alg_name: An algorithm to use for surrogate training during hyperparameter optimization
            to reduce number of fitness estimations with real ARTM model fitting.
            If not specifed, no surrogates will be used for hyper parameters search.
        :param surrogate_alg_params: A dict with the surrogate algorithm specific parameters.
            Depends on surrogate_alg_name. Should not be specified if surrogate_alg_name is not defined.
        :param artm_train_options: A dict with additional training options for underlying BigARTM implementation.
        :param working_dir_path: A directory where a nested temporary folder is created
            to store intermediate BigARTM files and other supplementary files.
            By default, working directory the current process is running with.
        :param texts_column_name: A name of the column in Pandas DataFrame to read texts of the dataset,
            if the dataset is represented as a 'pd.DataFrame'.
        :param log_file_path: A file path to log file for an optimization process.
        :param exp_id: An experiment id for an experiment versioning system (for example, Mlflow)
            to use for reporting final and intermediate results and metrics. Only Mlflow is currently supported.
        :param exp_tag: An experiment tag to log into Mlflow for later search purposes.
        :param exp_dataset_name: A dataset name to log into Mlflow for later search purposes.
        """
        self.topic_count = topic_count
        self.preprocessing_params = preprocessing_params
        self.alg_name = alg_name
        self.alg_params = alg_params
        self.surrogate_alg_name = surrogate_alg_name
        self.surrogate_alg_params = surrogate_alg_params
        self.artm_train_options = artm_train_options
        self.working_dir_path = working_dir_path
        self.texts_column_name = texts_column_name
        self.log_file_path = log_file_path
        self.exp_id = exp_id
        self.exp_tag = exp_tag
        self.exp_dataset_name = exp_dataset_name
        self._model: Optional[artm.ARTM] = None

    def fit(self, dataset: Union[pd.DataFrame, pd.Series], processed_dataset_path: Optional[str] = None) -> 'AutoTM':
        """
        Preprocess texts in the datasets, looks for the best hyperparameters for ARTM model and fits the model
        with these parameters. The instance will contain topics with the most probable words belonging to them.

        Parameters
        ----------
        dataset : DataFrame, Series containing texts of the corpus of size 'n_samples'.
            If dataset is DataFrame, the column containing texts will be identified
            from value of 'self.texts_column_name'.

        Returns
        -------
        self : object
            Fitted Estimator.
            :param processed_dataset_path: optional path where to write intermediate processed dataset

        """
        self._check_if_already_fitted(fit_is_ok=False)

        processed_dataset_path = processed_dataset_path or os.path.join(self.working_dir_path, f"{uuid.uuid4()}")

        logger.info(f"Stage 0: Create working dir {self.working_dir_path} if not exists")

        os.makedirs(self.working_dir_path, exist_ok=True)

        logger.info("Stage 1: Dataset preparation")
        # TODO: convert Series to DataFrame
        process_dataset(
            dataset,
            self.texts_column_name,
            processed_dataset_path,
            **self.preprocessing_params
        )
        prepare_all_artifacts(processed_dataset_path)
        logger.info("Stage 2: Tuning the topic model")

        if self.alg_name not in self._SUPPORTED_ALGS:
            raise ValueError(f"Alg {self.alg_name} is not supported. "
                             f"Only the following algorithms are supported: {self._SUPPORTED_ALGS}")

        if self.alg_name == "ga":
            # TODO: add checking of surrogate alg names
            # TODO: make mlflow arguments optional
            # exp_id and dataset_name will be needed further to store results in mlflow
            best_topic_model = genetic_algorithm.run_algorithm(
                data_path=processed_dataset_path,
                dataset=self.exp_dataset_name or "__noname__",
                exp_id=self.exp_id or "0",
                topic_count=self.topic_count,
                log_file=self.log_file_path,
                **(self.alg_params or dict())
            )
        else:
            # TODO: refactor this function
            best_topic_model = bayes_opt.run_algorithm(
                dataset=self.exp_dataset_name or "__noname__",
                data_path=processed_dataset_path,
                topic_count=self.topic_count,
                log_file=self.log_file_path,
                exp_id=self.exp_id or "0",
                **(self.alg_params or dict())
            )

        self._model = best_topic_model.model

        return self

    def predict(self, dataset: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
        """
        Looks for the best hyperparameters for ARTM model, fits the model with these parameters
        and predict topics mixtures for individual documents in the incoming corpus.

        Parameters
        ----------
        dataset : DataFrame or Series containing texts of the corpus of size 'n_samples'.
            If dataset is DataFrame, the column containing texts will be identified
            from value of 'self.texts_column_name'.

        Returns
        -------
        T : DataFrame of shape (n_samples, n_topics)
            Returns the probabilities of each topic to be in the every given text.
            Topic's probabilities are ordered according to topics ordering in 'self.topics' property.
        """
        self._check_if_already_fitted()

        os.makedirs(self.working_dir_path, exist_ok=True)

        with tempfile.TemporaryDirectory(dir=self.working_dir_path) as extractor_working_dir:
            if PROCESSED_TEXT_COLUMN not in dataset.columns:
                process_dataset(
                    dataset,
                    self.texts_column_name,
                    extractor_working_dir,
                    **self.preprocessing_params
                )
                preprocessed_dataset = pd.read_csv(os.path.join(extractor_working_dir, PREPOCESSED_DATASET_FILENAME))
            else:
                preprocessed_dataset = dataset
            topics_extractor = TopicsExtractor(self._model)
            mixtures = topics_extractor.get_prob_mixture(
                dataset=preprocessed_dataset, working_dir=extractor_working_dir
            )

        return mixtures

    def fit_predict(self, dataset: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
        """
        Preprocess texts in the datasets, looks for the best hyperparameters for ARTM model, fits the model
        with these parameters and predict topics mixtures for individual documents in the incoming corpus.
        The instance will contain topics with the most probable words belonging to them.

        Parameters
        ----------
        dataset : DataFrame or Series containing texts of the corpus of size 'n_samples'.
            If dataset is DataFrame, the column containing texts will be identified
            from value of 'self.texts_column_name'.

        Returns
        -------
        T : DataFrame of shape (n_samples, n_topics)
            Returns the probabilities of each topic to be in the every given text.
            Topic's probabilities are ordered according to topics ordering in 'self.topics' property.
        """
        self._check_if_already_fitted(fit_is_ok=False)

        processed_dataset_path = os.path.join(self.working_dir_path, f"{uuid.uuid4()}")
        self.fit(dataset, processed_dataset_path=processed_dataset_path)

        preprocessed_dataset = pd.read_csv(os.path.join(processed_dataset_path, PREPOCESSED_DATASET_FILENAME))
        return self.predict(preprocessed_dataset)

    def save(self, path: str, overwrite: bool = False):
        """
        Saves AutoTM to a filesystem.
        :param path: local filesystem path to save AutoTM on
        :param overwrite: if True and path: alredy exists, will try to remove path:
        """
        path_exists = os.path.exists(path)
        if path_exists and not overwrite:
            raise RuntimeError("The path is already exists and is not allowed to overwrite")
        elif path_exists:
            logger.debug(f"Removing existing path: {path}")
            shutil.rmtree(path)

        os.makedirs(path)

        artm_model_path = os.path.join(path, self._ARTM_MODEL_FILENAME)
        autotm_data_path = os.path.join(path, self._AUTOTM_DATA_FILENAME)

        self._model.dump_artm_model(artm_model_path)
        with open(autotm_data_path, "wb") as f:
            pickle.dump(self.get_params(), f)

    @property
    def topics(self) -> Dict[str, List[str]]:
        """
        Inferred set of topics with their respective top words.
        """
        self._check_if_already_fitted()
        return extract_topics(self._model)

    def print_topics(self):
        """
        Print topics in a human readable form in stdout
        """
        self._check_if_already_fitted()
        print_topics(self._model)

    def _check_if_already_fitted(self, fit_is_ok=True):
        if fit_is_ok and self._model is None:
            raise RuntimeError("The model is not fitted")

        if not fit_is_ok and self._model is not None:
            raise RuntimeError("The model is already fitted")



import io
import logging
import multiprocessing as mp
import os
import re
import sys
from collections import Counter
from contextlib import contextmanager
from datetime import datetime
from functools import partial
from multiprocessing import Pool
from typing import Dict, Any
from typing import Optional

import numpy as np
import pandas as pd

MetricsScores = Dict[str, Any]
TimeMeasurements = Dict[str, float]
AVG_COHERENCE_SCORE = "avg_coherence_score"
LLM_SCORE = "llm_score"

logger = logging.getLogger(__name__)


class TqdmToLogger(io.StringIO):
    """
    Output stream for TQDM which will output to logger module instead of
    the StdOut.
    """

    def __init__(self, base_logger, level=None):
        super(TqdmToLogger, self).__init__()
        self.logger = base_logger
        self.level = level or logging.INFO
        self.buf = ""

    def write(self, buf):
        self.buf = buf.strip("\r\n\t ")

    def flush(self):
        self.logger.log(self.level, self.buf)


class log_exec_timer:
    def __init__(self, name: Optional[str] = None):
        self.name = name
        self._start = None
        self._duration = None

    def __enter__(self):
        self._start = datetime.now()
        return self

    def __exit__(self, typ, value, traceback):
        self._duration = (datetime.now() - self._start).total_seconds()
        msg = (
            f"Exec time of {self.name}: {self._duration}"
            if self.name
            else f"Exec time: {self._duration}"
        )
        logger.info(msg)

    @property
    def duration(self):
        return self._duration


def merge_dicts(dicts):
    full_dict = {}
    for d in dicts:
        full_dict = dict(Counter(full_dict) + Counter(d))
    return full_dict


def parallelize_dataframe(df: pd.DataFrame, func, n_cores, return_type="df", **kwargs):
    """

    :param df: Dataframe to process.
    :param func: Function to be applied in parallel mode on data chunks
    :param n_cores: Amount of cores to parallelize on. In case of -1 takes all the available cores.
    :param return_type: datatype returned by func: 'df' or 'dict'
    :param kwargs: Additional parameters of the function, which is applied in parallel mode.
    :return: pd.DataFrame
    """
    if n_cores == -1:
        n_cores = mp.cpu_count() - 1
    df_split = np.array_split(df, n_cores)

    pool = Pool(n_cores)
    func_with_args = partial(func, **kwargs)
    map_res = pool.map(func_with_args, df_split)
    if return_type == "df":
        if isinstance(map_res[0], pd.DataFrame):
            res = pd.concat(map_res)
        elif isinstance(map_res[0], tuple):
            zipped_elems = list(zip(*map_res))
            res = (pd.concat(zipped_elems[0]), pd.concat(zipped_elems[1]))
    elif return_type == "dict":
        if isinstance(map_res[0], dict):
            res = merge_dicts(map_res)
        elif isinstance(map_res[0], tuple):
            zipped_elems = list(zip(*map_res))
            res = (merge_dicts(zipped_elems[0]), merge_dicts(zipped_elems[1]))
    else:
        raise ValueError(f"Unsupported return_type: {return_type}")
    pool.close()
    pool.join()
    return res


def make_log_config_dict(
        filename: str = "/var/log/tm-alg.txt", uid: Optional[str] = None,
        quiet: bool = False
) -> Dict[str, Any]:
    if filename is not None:
        if uid:
            dirname = os.path.dirname(filename)
            file, ext = os.path.splitext(os.path.basename(filename))
            log_filename = os.path.join(dirname, f"{file}-{uid}{ext}")
        else:
            log_filename = filename

        logfile_handler = {
            "logfile": {
                "level": "DEBUG",
                "formatter": "standard",
                "class": "logging.FileHandler",
                "filename": log_filename,
            }
        }
        handlers = ["logfile"]
        if not quiet:
            handlers.append("default")
    else:
        logfile_handler = dict()
        handlers = ["default"]

    return {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "standard": {
                "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            },
        },
        "handlers": {
            "default": {
                "level": "DEBUG",
                "formatter": "standard",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stderr",
            },
            **logfile_handler
        },
        "loggers": {
            "root": {
                "handlers": handlers,
                "level": "DEBUG",
                "propagate": False,
            },
            "GA": {
                "handlers": handlers,
                "level": "DEBUG",
                "propagate": False,
            },
            "GA_algo": {
                "handlers": handlers,
                "level": "DEBUG",
                "propagate": False,
            },
            "GA_surrogate": {
                "handlers": handlers,
                "level": "DEBUG",
                "propagate": False,
            },
        },
    }


@contextmanager
def do_suppress_stdout():
    save_stdout = sys.stdout
    save_stderr = sys.stderr
    stdout_file_name = "stdout.txt"
    stderr_file_name = "stderr.txt"
    try:
        with open(stdout_file_name, 'a') as out:
            with open(stderr_file_name, 'a') as err:
                sys.stdout = out
                sys.stderr = err
                yield
                out.flush()
                err.flush()
    finally:
        sys.stdout = save_stdout
        # sys.stderr = save_stderr
        if os.path.exists(stdout_file_name) and os.path.getsize(stdout_file_name) == 0:
            os.remove(stdout_file_name)
        if os.path.exists(stderr_file_name):
            with open(stderr_file_name, 'r') as file:
                content = file.read()
                if re.match(r'((^\d+it \[\d+:\d+, \?(\d+(\.\d+)?)?it/s\])?\n$)*', content):
                    os.remove(stderr_file_name)



import copy
import gc
import logging
import operator
import random
import sys
import time
import uuid
from typing import Optional, Callable, List

import numpy as np

from autotm.algorithms_for_tuning.genetic_algorithm.selection import selection
from autotm.algorithms_for_tuning.genetic_algorithm.statistics_collector import StatisticsCollector
from autotm.algorithms_for_tuning.individuals import IndividualDTO, Individual, IndividualBuilder
from autotm.algorithms_for_tuning.nelder_mead_optimization.nelder_mead import (
    NelderMeadOptimization,
)
from autotm.fitness.estimator import FitnessEstimator
from autotm.params import create_individual
from autotm.utils import AVG_COHERENCE_SCORE
from autotm.visualization.dynamic_tracker import MetricsCollector

ALG_ID = "ga"

logger = logging.getLogger("GA_algo")


def run_with_retry(action: Callable[[], object],
                   condition: Callable[[object], bool],
                   default_value: object = None,
                   max_retries: int = 5):
    for _ in range(max_retries):
        value = action()
        if condition(value):
            return value
    logger.warning(f"Cannot perform action after {max_retries} retries")
    return default_value


class ModelStorage:
    def __init__(self):
        self.stage_1_components = {}  # {config_id: id}
        self.stage_1_hyperp = {}  # {config_id: [[params1, params2]]}

    def model_search(self, model):
        raise NotImplementedError
        # for model.components


class GA:
    def __init__(
            self,
            dataset,
            data_path,
            num_individuals,
            num_iterations,
            ibuilder: IndividualBuilder,
            fitness_estimator: FitnessEstimator,
            mutation_type="mutation_one_param",
            crossover_type="blend_crossover",
            selection_type="fitness_prop",
            elem_cross_prob=0.2,
            early_stopping_iterations: Optional[int] = 500,
            best_proc=0.3,
            alpha=None,
            exp_id: Optional[int] = None,
            topic_count: Optional[int] = None,
            fitness_obj_type="single_objective",
            tag: Optional[str] = None,
            use_pipeline: bool = False,
            use_nelder_mead: bool = False,
            use_nelder_mead_in_mutation: bool = False,
            use_nelder_mead_in_crossover: bool = False,
            use_nelder_mead_in_selector: bool = False,
            train_option: str = "offline",
            statistics_collector: Optional[StatisticsCollector] = None,
            **kwargs,
    ):
        """

        :param dataset: dataset name
        :param data_path: path to data
        :param num_individuals: number of individuals
        :param num_iterations: number of iterations
        :param mutation_type: type of mutation, available types ['mutation_one_param', 'combined', 'psm', 'positioning_mutation']
        :param crossover_type: type of crossover, available types ['crossover_pmx', 'crossover_one_point', 'blend_crossover']
        :param selection_type: type of selection, available types ['fitness_prop', 'rank_based']
        :param elem_cross_prob: probability of crossover
        :param num_fitness_evaluations: number of fitness evaluations in case of limited resources
        :param early_stopping_iterations: number of iterations when there is no significant changes in fitness to stop training
        :param best_proc: percentage of parents to be transferred to new generation
        :param alpha:
        :param exp_id:
        :param surrogate_name:
        :param calc_scheme: how to apply surrogates, available values: 'type 1' - to use surrogates through iteration, 'type 2' - calculating 70% on each iteration
        :param topic_count:
        :param fitness_obj_type:
        :param tag:
        :param use_nelder_mead: should Nelder-Mead enchan
        :param kwargs:
        """

        self.dataset = dataset

        if crossover_type == "blend_crossover":
            self.crossover_children = 1
        else:
            self.crossover_children = 2
        self.data_path = data_path
        self.num_individuals = num_individuals
        self.num_iterations = num_iterations
        self.ibuilder = ibuilder
        self.fitness_estimator = fitness_estimator
        self.mutation_type = mutation_type
        self.crossover_type = crossover_type
        self.selection = selection(selection_type)
        self.elem_cross_prob = elem_cross_prob
        self.alpha = alpha
        self.early_stopping_iterations = early_stopping_iterations
        self.fitness_obj_type = fitness_obj_type
        self.best_proc = best_proc
        self.exp_id = exp_id
        self.topic_count = topic_count
        self.tag = tag
        self.use_pipeline = use_pipeline
        self.statistics_collector = statistics_collector
        # hyperparams
        self.use_nelder_mead = use_nelder_mead
        self.use_nelder_mead_in_mutation = use_nelder_mead_in_mutation
        self.use_nelder_mead_in_crossover = use_nelder_mead_in_crossover
        self.use_nelder_mead_in_selectior = use_nelder_mead_in_selector
        self.train_option = train_option
        self.metric_collector = MetricsCollector(
            dataset=self.dataset, n_specific_topics=topic_count
        )
        self.crossover_changes_dict = (
            {}
        )  # generation, parent_1_params, parent_2_params, ...

    def init_population(self):
        list_of_individuals = []
        for i in range(self.num_individuals):
            dto = IndividualDTO(
                id=str(uuid.uuid4()),
                data_path=self.data_path,
                dataset=self.dataset,
                fitness_name=self.ibuilder.individual_type,
                params=create_individual(base_model=i == 0, use_pipeline=self.use_pipeline),
                exp_id=self.exp_id,
                alg_id=ALG_ID,
                iteration_id=0,
                topic_count=self.topic_count,
                tag=self.tag,
                train_option=self.train_option,
            )
            # TODO: improve heuristic on search space
            list_of_individuals.append(self.ibuilder.make_individual(dto=dto))

        population_with_fitness = self.run_fitness(list_of_individuals, -1)

        # self.save_params(population_with_fitness)
        # if self.surrogate is not None and self.calc_scheme == "type2":
        #     self.surrogate.fit(np.array(self.all_params), np.array(self.all_fitness))
        #     logger.info("Surrogate is initialized!")

        self.fitness_estimator.fit(iter_num=-1)

        return population_with_fitness

    @staticmethod
    def _sort_population(population):
        population.sort(key=operator.attrgetter("fitness_value"), reverse=True)

    def run_crossover(self, pairs_generator, iteration_num: int):
        new_generation = []

        crossover_changes = {
            "parent_1_params": [],
            "parent_2_params": [],
            "parent_1_fitness": [],
            "parent_2_fitness": [],
            "child_id": [],
        }

        for i, j in pairs_generator:
            if i is None:
                break

            children = run_with_retry(
                action=lambda: i.params.crossover(j.params, crossover_type=self.crossover_type,
                                                  elem_cross_prob=self.elem_cross_prob, alpha=self.alpha),
                condition=lambda candidates: all(child.validate_params() for child in candidates),
                default_value=[]
            )

            children_dto = [IndividualDTO(
                id=str(uuid.uuid4()),
                data_path=self.data_path,
                dataset=self.dataset,
                params=child,
                exp_id=self.exp_id,
                alg_id=ALG_ID,
                iteration_id=iteration_num,
                topic_count=self.topic_count,
                tag=self.tag,
                train_option=self.train_option,
            ) for child in children]

            individuals = [self.ibuilder.make_individual(child) for child in children_dto]
            new_generation += individuals

            crossover_changes["parent_1_params"].append(i.params)
            crossover_changes["parent_2_params"].append(j.params)
            crossover_changes["parent_1_fitness"].append(i.fitness_value)
            crossover_changes["parent_2_fitness"].append(j.fitness_value)
            crossover_changes["child_id"].append(len(new_generation) - 1)

        if len(new_generation) > 0:
            new_generation = self.run_fitness(new_generation, iteration_num)
            logger.info(f"size of the new generation is {len(new_generation)}")

            for i in range(len(crossover_changes["parent_1_params"])):
                child_index = crossover_changes["child_id"][i]
                if child_index >= len(new_generation):
                    continue
                self.metric_collector.save_crossover(
                    generation=iteration_num,
                    parent_1=crossover_changes["parent_1_params"][i],
                    parent_2=crossover_changes["parent_2_params"][i],
                    parent_1_fitness=crossover_changes["parent_1_fitness"][i],
                    parent_2_fitness=crossover_changes["parent_2_fitness"][i],
                    child_1=new_generation[child_index].params,
                    child_1_fitness=new_generation[child_index].fitness_value,
                )

        return new_generation

    def apply_nelder_mead(self, starting_points_set, num_gen, num_iterations=2):
        nelder_opt = NelderMeadOptimization(
            data_path=self.data_path,
            dataset=self.dataset,
            exp_id=self.exp_id,
            topic_count=self.topic_count,
            train_option=self.train_option,
        )
        new_population = []
        for point in starting_points_set:
            st_point = point[:12] + [point[15]]
            res = nelder_opt.run_algorithm(
                num_iterations=num_iterations, ini_point=st_point
            )
            solution = list(res["x"])
            solution = (solution[:-1] + point[12:15] + [solution[-1]])  # TODO: check mutation ids
            fitness = -res.fun
            solution_dto = IndividualDTO(
                id=str(uuid.uuid4()),
                data_path=self.data_path,
                dataset=self.dataset,
                params=solution,
                exp_id=self.exp_id,
                alg_id=ALG_ID,
                iteration_id=num_gen,
                topic_count=self.topic_count,
                tag=self.tag,
                fitness_value={AVG_COHERENCE_SCORE: fitness},
                train_option=self.train_option,
            )

            new_population.append(self.ibuilder.make_individual(dto=solution_dto))
        return new_population

    def run(self, verbose=False, visualize_results=False) -> Individual:
        assert self.fitness_estimator.evaluations_counter == 0, \
            "Fitness estimator has non-zero evaluations count and cannot be reused"

        ftime = str(int(time.time()))

        # os.makedirs(LOG_FILE_PATH, exist_ok=True)

        logger.info(f"Starting experiment: {ftime}")

        logger.info(
            f"ALGORITHM PARAMS  number of individuals {self.num_individuals}; "
            f"number of fitness evals "
            f"{self.fitness_estimator.num_fitness_evaluations if self.fitness_estimator.num_fitness_evaluations else 'unlimited'}; "
            f"number of early stopping iterations "
            f"{self.early_stopping_iterations if self.early_stopping_iterations else 'unlimited'}; "
            f"crossover prob {self.elem_cross_prob}"
        )

        population = self.init_population()
        logger.info("POPULATION IS CREATED")

        x, y = [], []
        high_fitness = 0
        best_val_so_far = -10
        early_stopping_counter = 0

        run_id = str(uuid.uuid4())
        for ii in range(self.num_iterations):
            iteration_start_time = time.time()

            logger.info(f"ENTERING GENERATION {ii}")

            self._sort_population(population)
            if self.statistics_collector is not None:
                self.statistics_collector.log_iteration(
                    self.fitness_estimator.evaluations_counter,
                    population[0].fitness_value
                )
            pairs_generator = self.selection(
                population=population,
                best_proc=self.best_proc,
                children_num=self.crossover_children,
            )

            logger.info("PAIRS ARE CREATED")

            # Crossover
            new_generation = self.run_crossover(
                pairs_generator, iteration_num=ii
            )

            self._sort_population(new_generation)
            self._sort_population(population)

            logger.info("CROSSOVER IS OVER")

            if self.use_nelder_mead_in_crossover:
                # TODO: implement Nelder-Mead here
                pass

            del pairs_generator
            gc.collect()

            # population_params = [copy.deepcopy(individ.params) for individ in population]

            the_best_guy_params = copy.deepcopy(population[0].params)
            new_generation = [individ for individ in new_generation if individ.params != the_best_guy_params]

            new_generation_n = min(
                self.num_individuals - int(np.ceil(self.num_individuals * self.best_proc)),
                len(new_generation),
            )
            old_generation_n = self.num_individuals - new_generation_n

            population = population[:old_generation_n] + new_generation[:new_generation_n]

            try:
                del new_generation
            except Exception as e:
                logger.warning(e)
            gc.collect()

            self._sort_population(population)

            self.run_mutation(population)

            # after the mutation we obtain a new population that needs to be evaluated
            for p in population:
                p.dto.iteration_id = ii

            population = self.run_fitness(population, ii)

            # TODO (pipeline) Mutations collection is disabled
            # before_mutation = []  # individual
            # id_mutation = []  # new individual id
            # before_mutation.append(population[i])
            # id_mutation.append(i)
            # for ix, elem in enumerate(before_mutation):
            #     # TODO generation: int, original_params: list, mutated_params: list, original_fitness: float,
            #     #                       mutated_fitness: float
            #     self.metric_collector.save_mutation(
            #         generation=ii,
            #         original_params=elem.params,
            #         mutated_params=population[id_mutation[ix]].params,
            #         original_fitness=elem.fitness_value,
            #         mutated_fitness=population[id_mutation[ix]].fitness_value,
            #     )
            logger.info("MUTATION IS OVER")

            self._sort_population(population)

            if self.use_nelder_mead_in_mutation:
                collected_params = []
                for elem in population:
                    collected_params.append(elem.params)
                random_ids = random.sample(
                    [i for i in range(len(collected_params))], k=3
                )
                starting_points = [collected_params[i] for i in random_ids]

                nm_population = self.apply_nelder_mead(starting_points, num_gen=ii)
                for i, elem in enumerate(nm_population):
                    if population[i].fitness_value < elem.fitness_value:
                        logging.info(
                            f"NM found better solution! {elem.fitness_value} vs {population[i].fitness_value}"
                        )
                        population[i] = elem

            if self.fitness_estimator.num_fitness_evaluations and self.fitness_estimator.evaluations_counter >= self.fitness_estimator.num_fitness_evaluations:
                self.metric_collector.save_fitness(
                    generation=ii,
                    params=[i.params for i in population],
                    fitness=[i.fitness_value for i in population],
                )
                bparams = "".join([str(i) for i in population[0].params])
                logger.info(
                    f"TERMINATION IS TRIGGERED: EVAL NUM (2)."
                    f"DATASET {self.dataset}."
                    f"TOPICS NUM {self.topic_count}."
                    f"RUN ID {run_id}."
                    f"THE BEST FITNESS {population[0].fitness_value}."
                    f"THE BEST PARAMS {bparams}."
                    f"ITERATION TIME {time.time() - iteration_start_time}."
                )
                break

            current_fitness = population[0].fitness_value
            if (current_fitness > high_fitness) or (ii == 0):
                high_fitness = current_fitness

            # if self.surrogate:
            #     if self.calc_scheme == "type1" and not surrogate_iteration:
            #         self.surrogate.fit(
            #             np.array(self.all_params), np.array(self.all_fitness)
            #         )
            #     elif self.calc_scheme == "type2":
            #         self.surrogate.fit(
            #             np.array(self.all_params), np.array(self.all_fitness)
            #         )

            self.fitness_estimator.fit(ii)

            if self.early_stopping_iterations:
                if population[0].fitness_value > best_val_so_far:
                    best_val_so_far = population[0].fitness_value
                    early_stopping_counter = 0
                else:
                    early_stopping_counter += 1
                    if early_stopping_counter == self.early_stopping_iterations:
                        bparams = "".join([str(i) for i in population[0].params])
                        self.metric_collector.save_fitness(
                            generation=ii,
                            params=[i.params for i in population],
                            fitness=[i.fitness_value for i in population],
                        )
                        logger.info(
                            f"TERMINATION IS TRIGGERED: EARLY STOPPING."
                            f"DATASET {self.dataset}."
                            f"TOPICS NUM {self.topic_count}."
                            f"RUN ID {run_id}."
                            f"THE BEST FITNESS {population[0].fitness_value}."
                            f"THE BEST PARAMS {bparams}."
                            f"ITERATION TIME {time.time() - iteration_start_time}."
                        )
                        break

            x.append(ii)
            y.append(population[0].fitness_value)
            self.metric_collector.save_fitness(
                generation=ii,
                params=[i.params for i in population],
                fitness=[i.fitness_value for i in population],
            )
            logger.info(
                f"Population len {len(population)}. "
                f"Best params so far: {population[0].params}, with fitness: {population[0].fitness_value}."
                f"ITERATION TIME: {time.time() - iteration_start_time}"
                f"DATASET {self.dataset}."
                f"TOPICS NUM {self.topic_count}."
                f"RUN ID {run_id}."
            )
            best_solution = population[0]
            self.fitness_estimator.log_best_solution(best_solution, alg_args=" ".join(sys.argv), is_tmp=True)

        if visualize_results:
            self.metric_collector.save_and_visualise_trace()
        else:
            self.metric_collector.save_trace()

        if self.statistics_collector is not None:
            self.statistics_collector.log_iteration(
                self.fitness_estimator.evaluations_counter,
                population[0].fitness_value
            )
        logger.info(f"Y: {y}")
        best_individual = population[0]
        ind = self.fitness_estimator.log_best_solution(best_individual, alg_args=" ".join(sys.argv))
        logger.info(f"Logged the best solution. Obtained fitness is {ind.fitness_value}")

        return ind

    def run_fitness(self, population: List[Individual], ii: int):
        return self.fitness_estimator.estimate(ii,population )

    def run_mutation(self, population):
        for i in range(1, len(population)):
            if random.random() <= population[i].params.mutation_probability:
                params = run_with_retry(
                    action=lambda: population[i].params.mutate(mutation_type=self.mutation_type),
                    condition=lambda candidate: candidate.validate_params()
                )
                if params is None:
                    continue

                dto = IndividualDTO(
                    id=str(uuid.uuid4()),
                    data_path=self.data_path,
                    dataset=self.dataset,
                    params=params,
                    exp_id=self.exp_id,
                    alg_id=ALG_ID,
                    topic_count=self.topic_count,
                    tag=self.tag,
                    train_option=self.train_option,
                )
                population[i] = self.ibuilder.make_individual(dto=dto)


# multistage bag of regularizers approach
# TODO: add penalty for the solution length


class GAmultistage(GA):
    def __init__(self, dataset, num_individuals, max_stages=5):  # max_stage_len
        self.max_stages = max_stages  # amount of unique regularizers
        self.dataset = dataset
        self.bag_of_regularizers = [
            "decor_S",
            "decor_B",
            "S_phi_B",
            "S_phi_S",
            "S_theta_B",
            "S_theta_S",
        ]  # add separate smooth and sparsity
        self.num_individuals = num_individuals
        self.initial_element_stage_probability = 0.5
        self.positioning_matrix = np.full(
            (len(self.bag_of_regularizers), self.max_stages - 1), 0.5
        )
        self.set_regularizer_limits()

    def set_regularizer_limits(
        self,
        low_decor=0,
        high_decor=1e5,
        low_n=1,
        high_n=30,  # minimal value changed to 1
        low_back=0,
        high_back=5,
        low_spb=0,
        high_spb=1e2,
        low_spm=-1e-3,
        high_spm=1e2,
        low_sp_phi=-1e3,
        high_sp_phi=1e3,
    ):
        self.high_decor = high_decor
        self.low_decor = low_decor
        self.low_n = low_n
        self.high_n = high_n
        self.low_back = low_back
        self.high_back = high_back
        self.high_spb = high_spb
        self.low_spb = low_spb
        self.low_spm = low_spm
        self.high_spm = high_spm
        self.low_sp_phi = low_sp_phi
        self.high_sp_phi = high_sp_phi

    # TODO: check if float is needed
    def _init_param(self, param_type):
        if param_type == "decor_S" or param_type == "decor_B":
            return float(
                np.random.uniform(low=self.low_decor, high=self.high_decor, size=1)[0]
            )
        elif param_type == "S_phi_B" or "S_theta_B":
            return float(
                np.random.uniform(low=self.low_spb, high=self.high_spb, size=1)[0]
            )
        elif param_type == "S_phi_S" or "S_theta_S":
            return float(
                np.random.uniform(low=self.low_sp_phi, high=self.high_sp_phi, size=1)[0]
            )
        elif param_type == "n":
            return float(np.random.randint(low=self.low_n, high=self.high_n, size=1)[0])
        elif param_type == "B":
            return float(
                np.random.randint(low=self.low_back, high=self.high_back, size=1)[0]
            )

    def _create_stage(self, stage_num):
        stage_regularizers = []
        for ix, elem in enumerate(self.bag_of_regularizers):
            elem_sample_prob = self.positioning_matrix[ix][stage_num - 1]
            if random.random() < elem_sample_prob:
                stage_regularizers.append(elem)
        stage_regularizers.append("n")
        return stage_regularizers

    def init_individ(self):
        number_of_stages = np.random.randint(low=1, high=self.max_stages, size=1)[0]
        dict_of_stages = [{} for _ in range(number_of_stages)]
        for i in range(number_of_stages):
            regularizers = self._create_stage(i)
            for reg_name in regularizers:
                value = self._init_param(reg_name)
                dict_of_stages[i][reg_name] = value
        dict_of_stages = [{"B": self._init_param("B")}] + dict_of_stages
        return dict_of_stages

    def init_population(self):
        raise NotImplementedError



import logging
import os
import time
from multiprocessing.process import current_process
from typing import List, Optional, cast

import celery
from billiard.exceptions import SoftTimeLimitExceeded
from celery import shared_task, group, Task
from celery.exceptions import TimeLimitExceeded
from celery.result import GroupResult, AsyncResult
from celery.utils.log import get_task_logger
from tqdm import tqdm

from autotm.algorithms_for_tuning.individuals import Individual, IndividualBuilder
from autotm.fitness.tm import fit_tm_of_individual
from autotm.params_logging_utils import model_files, log_params_and_artifacts, log_stats
from autotm.schemas import IndividualDTO
from autotm.utils import TqdmToLogger

logger = logging.getLogger("root")
task_logger = get_task_logger(__name__)


class RemotelyCalculableIndividual:
    def get_params(self) -> str:
        raise NotImplementedError()

    def set_fitness(self, value):
        raise NotImplementedError()


def make_celery_app(client_to_broker_max_retries=3):
    app = celery.Celery(
        "distributed_fitness",
        backend=os.environ.get("CELERY_RESULT_BACKEND"),
        broker=os.environ["CELERY_BROKER_URL"]
    )

    # for explanation of this options see:
    # 1. https://stackoverflow.com/questions/56805193/how-to-set-up-celery-producer-send-task-timeout
    # 2. https://github.com/celery/celery/issues/4627#issuecomment-396907957
    app.conf.update(
        task_acks_late=True,
        task_reject_on_worker_lost=True,
        task_acks_on_failure_or_timeout=False,
        worker_prefetch_multiplier=1,
        broker_transport_options={
            "max_retries": client_to_broker_max_retries,
            "interval_start": 0,
            "interval_step": 0.2,
            "interval_max": 0.5
        }
    )

    return app


def do_fitness_calculating(individual: str,
                           log_artifact_and_parameters: bool = False,
                           log_run_stats: bool = False,
                           alg_args: Optional[str] = None) -> str:
    individual: IndividualDTO = IndividualDTO.parse_raw(individual)
    task_logger.info(f"Calculating fitness for an individual with id {individual.id}: \n{individual}")

    with fit_tm_of_individual(
            dataset=individual.dataset,
            data_path=individual.data_path,
            params=individual.params,
            fitness_name=individual.fitness_name,
            topic_count=individual.topic_count,
            force_dataset_settings_checkout=individual.force_dataset_settings_checkout,
            train_option=individual.train_option) as (time_metrics, metrics, tm):
        individual.fitness_value = metrics

        with model_files(tm) as tm_files:
            if log_artifact_and_parameters:
                log_params_and_artifacts(tm, tm_files, individual, time_metrics, alg_args)

            if log_run_stats:
                log_stats(tm, tm_files, individual, time_metrics, alg_args)

    task_logger.info(f"Fitness has been calculated for {individual.id}: {individual.fitness_value}")

    return individual.json()


# @shared_task(time_limit=25*60, soft_time_limit=20*60, autoretry_for=(Exception,),
# retry_kwargs={'max_retries': 3, 'countdown': 5})
@shared_task(bind=True, time_limit=25*60, soft_time_limit=20*60)
def calculate_fitness(self: Task,
                      individual: str,
                      log_artifact_and_parameters: bool = False,
                      log_run_stats: bool = False,
                      alg_args: Optional[str] = None) -> str:
    # A hack to solve the problem 'AssertionError: daemonic processes are not allowed to have children'
    # see the discussion: https://github.com/celery/celery/issues/1709
    # noinspection PyUnresolvedReferences
    current_process()._config['daemon'] = False

    try:
        return do_fitness_calculating(individual, log_artifact_and_parameters, log_run_stats, alg_args)
    except TimeLimitExceeded as ex:
        logger.error("Encountered hard time limit!")
    except SoftTimeLimitExceeded as ex:
        raise Exception(f"Soft time limit encountered") from ex
    except Exception:
        self.retry(max_retries=1, countdown=5)


def parallel_fitness(ibuilder: IndividualBuilder,
                     population: List[Individual],
                     use_tqdm: bool = False,
                     tqdm_check_period: int = 2,
                     app: Optional[celery.Celery] = None) -> List[Individual]:
    ids = [ind.dto.id for ind in population]
    assert len(set(ids)) == len(population), \
        f"There are individuals with duplicate ids: {ids}"

    logger.info("Calculating fitness...")
    logger.info(f"Sending individuals to be calculated with uids: {[p.dto.id for p in population]}")

    fitness_tasks = []
    for individual in population:
        task = cast(
            Task,
            calculate_fitness.signature((individual.dto.json(), False, False), options={"queue": "fitness_tasks"})
        )
        # todo: add it later
        # if app is not None:
        #     task.bind(app)
        fitness_tasks.append(task)

    g: GroupResult = group(*fitness_tasks).apply_async()

    logger.info(f"Corresponding celery tasks ids: {[child.id for child in g.children]}")

    tqdm_out = TqdmToLogger(logger, level=logging.INFO)

    if use_tqdm:
        # TODO: add timeout here
        with tqdm(total=len(fitness_tasks), file=tqdm_out) as pbar:
            while True:
                cc = g.completed_count()
                # logger.debug(f"Completed task count : {cc}")
                pbar.update(cc - pbar.n)

                # logger.debug(f"is ready {g.ready()} or failed {g.failed()}")

                if g.ready() or g.failed():
                    break

                # TODO: make it into parameter
                time.sleep(tqdm_check_period)

    # TODO: ugly workround to solve indefinite hanging g.get(), see pages below for additional info
    # https://stackoverflow.com/questions/63860955/celery-async-result-get-hangs-waiting-for-result-even-after-celery-worker-has
    # https://stackoverflow.com/questions/49006182/celery-redis-get-hangs-indefinitely-after-running-smoothly-for-70-hours
    logger.info("Getting results")
    # TODO: this is commented because of https://github.com/celery/celery/pull/7040
    # waiting_delay = int(os.environ.get('AUTOTM_KUBE_FITNESS_WAITING_DELAY', '90'))
    # while not g.ready():
    #     logger.info(f"Results are not ready. Waiting for the next {waiting_delay} seconds...")
    #     time.sleep(waiting_delay)
    # logger.info("Results are ready, trying ot obtain them...")
    # results = g.get()
    results = g.get()
    logger.info("The results have been obtained")

    # restoring the order in the resulting population according to the initial population
    # results_by_id = {ind.id: ind for ind in (fitness_from_json(r) for r in results)}
    results_by_id = {ind.id: ind for ind in (IndividualDTO.parse_raw(r) for r in results)}
    return [ibuilder.make_individual(results_by_id[ind.dto.id]) for ind in population]


def log_best_solution(ibuilder: IndividualBuilder,
                      individual: Individual,
                      wait_for_result_timeout: Optional[float] = None,
                      alg_args: Optional[str] = None,
                      is_tmp: bool = False,
                      app: Optional[celery.Celery] = None) \
        -> Individual:
    if is_tmp:
        return ibuilder.make_individual(individual.dto)
    # ind = fitness_to_json(individual)
    ind = individual.dto.json()
    logger.info(f"Sending a best individual to be logged: {ind}")
    task: Task = calculate_fitness.signature(
        (ind, True, False, alg_args),
        options={"queue": "fitness_tasks"}
    )

    # todo: add it later
    # if app is not None:
    #     task.bind(app)

    result: AsyncResult = task.apply_async()
    logger.debug(f"Started a task for best individual logging with id: {result.task_id}")

    # if wait_for_result_timeout:
    #     # it may block forever
    #     timeout = wait_for_result_timeout if wait_for_result_timeout > 0 else None
    #     r = result.get(timeout=timeout)
    #     # r = fitness_from_json(r)
    #     r = IndividualDTO.parse_raw(r)
    # else:
    #     r = result

    r = result.get(timeout=wait_for_result_timeout)
    r = IndividualDTO.parse_raw(r)
    ind = ibuilder.make_individual(r)

    return ind



from typing import Optional
from typing import Union

from pydantic import BaseModel

from autotm.params import PipelineParams, FixedListParams
from autotm.utils import MetricsScores

AnyParams = Union[FixedListParams, PipelineParams]


class IndividualDTO(BaseModel):
    id: str
    data_path: str
    params: AnyParams
    fitness_name: str = "regular"
    dataset: str = "default"
    force_dataset_settings_checkout: bool = False
    fitness_value: Optional[MetricsScores] = None
    exp_id: Optional[int] = None
    alg_id: Optional[str] = None
    tag: Optional[str] = None
    iteration_id: int = 0
    topic_count: Optional[int] = None
    train_option: str = "offline"

    class Config:
        arbitrary_types_allowed = True

    def make_params_dict(self):
        return self.params.make_params_dict()




import copy
import logging
import math
import time
import uuid
from abc import ABC, abstractmethod
from typing import List, Optional

import numpy as np

from autotm.abstract_params import AbstractParams
from autotm.algorithms_for_tuning.genetic_algorithm.statistics_collector import StatisticsCollector
from autotm.algorithms_for_tuning.genetic_algorithm.surrogate import Surrogate, set_surrogate_fitness, \
    get_prediction_uncertanty
from autotm.algorithms_for_tuning.individuals import Individual, IndividualBuilder
from autotm.fitness import local_tasks, cluster_tasks
from autotm.schemas import IndividualDTO

logger = logging.getLogger(__name__)


class FitnessEstimator:
    def __init__(self, num_fitness_evaluations: Optional[int] = None, statistics_collector: Optional[StatisticsCollector] = None):
        self._num_fitness_evaluations = num_fitness_evaluations
        self._evaluations_counter = 0
        self._statistics_collector = statistics_collector
        super().__init__()

    @property
    def num_fitness_evaluations(self) -> Optional[int]:
        return self._num_fitness_evaluations

    @property
    def evaluations_counter(self) -> int:
        return self._evaluations_counter

    @abstractmethod
    def fit(self, iter_num: int) -> None:
        ...

    @abstractmethod
    def log_best_solution(self,
                          individual: Individual,
                          wait_for_result_timeout: Optional[float] = None,
                          alg_args: Optional[str] = None,
                          is_tmp: bool = False) -> Individual:
        ...

    def estimate(self, iter_num: int, population: List[Individual]) -> List[Individual]:
        evaluated = [individual for individual in population if individual.dto.fitness_value is not None]
        not_evaluated = [individual for individual in population if individual.dto.fitness_value is None]
        evaluations_limit = max(0, self._num_fitness_evaluations - self._evaluations_counter) \
            if self._num_fitness_evaluations else len(not_evaluated)
        if len(not_evaluated) > evaluations_limit:
            not_evaluated = not_evaluated[:evaluations_limit]
        self._evaluations_counter += len(not_evaluated)
        new_evaluated = self._estimate(iter_num, not_evaluated)
        if self._statistics_collector:
            for individual in new_evaluated:
                self._statistics_collector.log_individual(individual)
        return evaluated + new_evaluated

    @abstractmethod
    def _estimate(self, iter_num: int, population: List[Individual]) -> List[Individual]:
        ...


class SurrogateEnabledFitnessEstimatorMixin(FitnessEstimator, ABC):
    SUPPORTED_CALC_SCHEMES = ["type1", "type2"]

    ibuilder: IndividualBuilder
    surrogate: Surrogate
    calc_scheme: str
    speedup: bool
    all_params: List[AbstractParams]
    all_fitness: List[float]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def surrogate_iteration(iter_num: int) -> bool:
        return (iter_num % 2 != 0) if iter_num > 0 else False

    def fit(self, iter_num: int) -> None:
        surrogate_iteration = self.surrogate_iteration(iter_num)
        if (self.calc_scheme == "type1" and not surrogate_iteration) or (self.calc_scheme == "type2"):
            self.surrogate.fit(np.array(self.all_params), np.array(self.all_fitness))

    def _estimate(self, iter_num: int, population: List[Individual]) -> List[Individual]:
        fitness_calc_time_start = time.time()
        surrogate_iteration = self.surrogate_iteration(iter_num)

        if not self.speedup or not surrogate_iteration or iter_num == -1:
            population = super().estimate(iter_num, population)
            self.save_params(population)

        if self.calc_scheme == "type1" and surrogate_iteration:
            population = self.surrogate_calculation(population)
        elif self.calc_scheme == "type2" and iter_num != -1:
            population = self._calculate_uncertain_res(iter_num, population)
            self.save_params(population)

        logger.info(f"TIME OF THE SURROGATE-BASED FITNESS FUNCTION: {time.time() - fitness_calc_time_start}")

        return population

    def surrogate_calculation(self, population: List[Individual]):
        x_val = np.array([copy.deepcopy(individ.params.to_vector()) for individ in population])
        y_pred = self.surrogate.predict(x_val)

        if not self.speedup:
            y_val = np.array([individ.fitness_value for individ in population])

            def check_val(fval):
                return not (fval is None or math.isnan(fval) or math.isinf(fval))

            def check_params(p):
                return all(check_val(el) for el in p)

            clean_params_and_f = []
            for i in range(len(y_val)):
                if not check_params(x_val[i]) or not check_val(y_val[i]):
                    logger.warning(
                        f"Bad params or fitness found. Fitness: {y_val[i]}. Params: {x_val[i]}."
                    )
                else:
                    clean_params_and_f.append((x_val[i], y_val[i]))

            x_val = clean_params_and_f[0]
            y_val = clean_params_and_f[1]
            r_2, mse, rmse = self.surrogate.score(x_val, y_val)
            logger.info(f"Real values: {list(y_val)}")
            logger.info(f"Predicted values: {list(y_pred)}")
            logger.info(f"R^2: {r_2}, MSE: {mse}, RMSE: {rmse}")

        for ix, individ in enumerate(population):
            individ.dto.fitness_value = set_surrogate_fitness(y_pred[ix])

        return population

    def _calculate_uncertain_res(self, iter_num: int, population: List[Individual], proc:float = 0.3):
        if len(population) == 0:
            return []

        x = np.array([individ.dto.params.to_vector() for individ in population])
        certanty = get_prediction_uncertanty(
            self.surrogate.surrogate, x, self.surrogate.name
        )
        recalculate_num = int(np.floor(len(certanty) * proc))
        logger.info(f"Certanty values: {certanty}")

        certanty, x = (
            list(t) for t in zip(*sorted(zip(certanty, x.tolist()), reverse=True))
        )  # check
        calculated = []
        for individual in population[:recalculate_num]:
            # copy
            individual_json = individual.dto.model_dump_json()
            individual = self.ibuilder.make_individual(dto=IndividualDTO.model_validate_json(individual_json))
            individual.dto.fitness_value = None
            calculated.append(individual)

        calculated = super().estimate(iter_num, calculated)

        self.all_params += [individ.dto.params.to_vector() for individ in calculated]
        self.all_fitness += [
            individ.dto.fitness_value["avg_coherence_score"] for individ in calculated
        ]

        pred_y = self.surrogate.predict(x[recalculate_num:])
        for ix, individual in enumerate(population[recalculate_num:]):
            dto = individual.dto
            dto = IndividualDTO(
                id=str(uuid.uuid4()),
                data_path=dto.data_path,
                params=dto.params,
                dataset=dto.dataset,
                fitness_value=set_surrogate_fitness(pred_y[ix]),
                exp_id=dto.exp_id,
                alg_id=dto.alg_id,
                topic_count=dto.topic_count,
                tag=dto.tag,
                train_option=dto.train_option,
            )
            calculated.append(self.ibuilder.make_individual(dto=dto))
        return calculated

    def save_params(self, population):
        params_and_f = [
            (copy.deepcopy(individ.params.to_vector()), individ.fitness_value)
            for individ in population
            if individ.fitness_value not in self.all_fitness
        ]

        def check_val(fval):
            return not (fval is None or math.isnan(fval) or math.isinf(fval))

        def check_params(pp):
            return all(check_val(el) for el in pp)

        clean_params_and_f = []
        for p, f in params_and_f:
            if not check_params(p) or not check_val(f):
                logger.warning(f"Bad params or fitness found. Fitness: {f}. Params: {p}.")
            else:
                clean_params_and_f.append((p, f))

        pops = [p for p, _ in clean_params_and_f]
        fs = [f for _, f in clean_params_and_f]

        self.all_params += pops
        self.all_fitness += fs


class ComputableFitnessEstimator(FitnessEstimator):
    def __init__(self,
                 ibuilder: IndividualBuilder,
                 num_fitness_evaluations: Optional[int] = None,
                 statistics_collector: Optional[StatisticsCollector] = None):
        self.ibuilder = ibuilder
        super().__init__(num_fitness_evaluations, statistics_collector)

    def fit(self, iter_num: int) -> None:
        pass

    def log_best_solution(self,
                          individual: Individual,
                          wait_for_result_timeout: Optional[float] = None,
                          alg_args: Optional[str] = None,
                          is_tmp: bool = False) -> Individual:
        return local_tasks.log_best_solution(self.ibuilder, individual, wait_for_result_timeout, alg_args, is_tmp)

    def _estimate(self, iter_num: int, population: List[Individual]) -> List[Individual]:
        return local_tasks.estimate_fitness(self.ibuilder, population)


class DistributedComputableFitnessEstimator(FitnessEstimator):
    def __init__(self,
                 ibuilder: IndividualBuilder,
                 num_fitness_evaluations: Optional[int] = None,
                 statistics_collector: Optional[StatisticsCollector] = None):
        self.app = cluster_tasks.make_celery_app()
        self.ibuilder = ibuilder
        super().__init__(num_fitness_evaluations, statistics_collector)

    def fit(self, iter_num: int) -> None:
        pass

    def log_best_solution(self,
                          individual: Individual,
                          wait_for_result_timeout: Optional[float] = None,
                          alg_args: Optional[str] = None,
                          is_tmp: bool = False) -> Individual:
        return cluster_tasks.log_best_solution(self.ibuilder, individual,
                                               wait_for_result_timeout, alg_args, is_tmp, app=self.app)

    def _estimate(self, iter_num: int, population: List[Individual]) -> List[Individual]:
        return cluster_tasks.parallel_fitness(self.ibuilder, population, app=self.app)


class SurrogateEnabledComputableFitnessEstimator(ComputableFitnessEstimator, SurrogateEnabledFitnessEstimatorMixin):
    def __init__(self,
                 ibuilder: IndividualBuilder,
                 surrogate: Surrogate,
                 calc_scheme: str,
                 speedup: bool = True,
                 num_fitness_evaluations: Optional[int] = None,
                 statistics_collector: Optional[StatisticsCollector] = None):
        self.ibuilder = ibuilder
        self.surrogate = surrogate
        self.calc_scheme = calc_scheme
        self.speedup = speedup

        self.all_params: List[AbstractParams] = []
        self.all_fitness: List[float] = []

        if calc_scheme not in self.SUPPORTED_CALC_SCHEMES:
            raise ValueError(f"Unexpected surrogate scheme! {self.calc_scheme}")
        super().__init__(ibuilder, num_fitness_evaluations, statistics_collector)


class DistributedSurrogateEnabledComputableFitnessEstimator(
    DistributedComputableFitnessEstimator,
    SurrogateEnabledFitnessEstimatorMixin
):
    def __init__(self,
                 ibuilder: IndividualBuilder,
                 surrogate: Surrogate,
                 calc_scheme: str,
                 speedup: bool = True,
                 num_fitness_evaluations: Optional[int] = None,
                 statistics_collector: Optional[StatisticsCollector] = None):
        self.ibuilder = ibuilder
        self.surrogate = surrogate
        self.calc_scheme = calc_scheme
        self.speedup = speedup

        self.all_params: List[AbstractParams] = []
        self.all_fitness: List[float] = []

        if calc_scheme not in self.SUPPORTED_CALC_SCHEMES:
            raise ValueError(f"Unexpected surrogate scheme! {self.calc_scheme}")
        super().__init__(ibuilder, num_fitness_evaluations, statistics_collector)



import random
from typing import List

import numpy as np


def mutation_one_param(
        individ: List[float],
        low_spb: float,
        high_spb: float,
        low_spm: float,
        high_spm: float,
        low_n: int,
        high_n: int,
        low_back: int,
        high_back: int,
        low_decor: float,
        high_decor: float,
        elem_mutation_prob: float = 0.1,
):
    """
    One-point mutation

    Checking the probability of mutation for each of the elements

    Parameters
    ----------
    individ: List[float]
        Individual to be processed
    low_spb: float
        The lower possible bound for sparsity regularizer of back topics
    high_spb: float
        The higher possible bound for sparsity regularizer of back topics
    low_spm: float
        The lower possible bound for sparsity regularizer of specific topics
    high_spm: float
        The higher possible bound for sparsity regularizer of specific topics
    low_n: int
        The lower possible bound for amount of iterations between stages
    high_n: int
        The higher possible bound for amount of iterations between stages
    low_back:
        The lower possible bound for amount of back topics
    high_back:
        The higher possible bound for amount of back topics


    Returns
    ----------
    Updated individuals with exchanged chromosome parts
    """
    for i in range(len(individ)):
        if random.random() <= elem_mutation_prob:
            if i in [2, 3]:
                individ[i] = np.random.uniform(low=low_spb, high=high_spb, size=1)[0]
            for i in [5, 6, 8, 9]:
                individ[i] = np.random.uniform(low=low_spm, high=high_spm, size=1)[0]
            for i in [1, 4, 7, 10]:
                individ[i] = float(np.random.randint(low=low_n, high=high_n, size=1)[0])
            for i in [11]:
                individ[i] = float(np.random.randint(low=low_back, high=high_back, size=1)[0])
            for i in [0, 15]:
                individ[i] = np.random.uniform(low=low_decor, high=high_decor, size=1)[0]
    return individ


def positioning_mutation(individ, elem_mutation_prob=0.1, **kwargs):
    for i in range(len(individ)):
        set_1 = set([i])
        if random.random() <= elem_mutation_prob:
            if i in [0, 15]:
                set_2 = set([0, 15])
                ix = np.random.choice(list(set_2.difference(set_1)))
                tmp = individ[ix]
                individ[ix] = individ[i]
                individ[i] = tmp
            elif i in [1, 4, 7, 10, 11]:
                set_2 = set([1, 4, 7, 10, 11])
                ix = np.random.choice(list(set_2.difference(set_1)))
                tmp = individ[ix]
                individ[ix] = individ[i]
                individ[i] = tmp
            elif i in [2, 3]:
                set_2 = set([2, 3])
                ix = np.random.choice(list(set_2.difference(set_1)))
                tmp = individ[ix]
                individ[ix] = individ[i]
                individ[i] = tmp
            elif i in [5, 6, 8, 9]:
                set_2 = set([5, 6, 8, 9])
                ix = np.random.choice(list(set_2.difference(set_1)))
                tmp = individ[ix]
                individ[ix] = individ[i]
                individ[i] = tmp
        return individ


def mutation_combined(individ, elem_mutation_prob=0.1, **kwargs):
    if random.random() <= individ[14]:  # TODO: check 14th position
        return mutation_one_param(individ, elem_mutation_prob)
    else:
        return positioning_mutation(individ, elem_mutation_prob)
    pass


def do_swap_in_ranges(individ, i, ranges):
    swap_range = next((r for r in ranges if i in r), None)
    if swap_range is not None:
        swap_range = [j for j in swap_range if i != j]
        j = np.random.choice(swap_range)
        individ[i], individ[j] = individ[j], individ[i]


PSM_NEW_SWAP_RANGES = [[1, 4, 7, 10], [2, 3], [5, 6, 8, 9]]


def mutation_psm_new(individ, elem_mutation_prob, **kwargs):
    for i in range(len(individ)):
        if random.random() < elem_mutation_prob:
            if i == 0:
                individ[i] = individ[15]
            elif i == 15:
                individ[i] = individ[0]
            else:
                do_swap_in_ranges(individ, i, PSM_NEW_SWAP_RANGES)
    return individ


PSM_SWAP_RANGES = [[1, 4, 7], [2, 5], [3, 6]]


def mutation_psm(individ, elem_mutation_prob=0.1, **kwargs):
    for i in range(len(individ)):
        if random.random() < elem_mutation_prob:
            if i == 0:
                individ[i] = np.random.uniform(low=1, high=100, size=1)[0]
            else:
                do_swap_in_ranges(individ, i, PSM_SWAP_RANGES)
    return individ


def mutation(mutation_type="mutation_one_param"):
    if mutation_type == "mutation_one_param":
        return mutation_one_param
    if mutation_type == "combined":
        return mutation_combined
    if mutation_type == "psm":
        return mutation_psm
    if mutation_type == "positioning_mutation":
        return positioning_mutation



import numpy as np


# additional scores to calculate


# Topic Significance
def kl_divergence(p, q):
    return np.sum(np.where(p != 0, p * np.log(p / q), 0))


# Uniform Distribution Over Words (W-Uniform)
def ts_uniform(topic_word_dist):
    n_words = topic_word_dist.shape[0]
    w_uniform = np.ones(n_words) / n_words
    uniform_distances_kl = [kl_divergence(p, w_uniform) for p in topic_word_dist.T]
    return uniform_distances_kl


# Vacuous Semantic Distribution (W-Vacuous)
def ts_vacuous(doc_topic_dist, topic_word_dist, total_tokens):
    n_words = topic_word_dist.shape[0]
    #     n_tokens = np.sum([len(text) for text in texts])
    p_k = np.sum(doc_topic_dist, axis=1) / total_tokens
    w_vacauous = np.sum(topic_word_dist * np.tile(p_k, (n_words, 1)), axis=1)
    vacauous_distances_kl = [kl_divergence(p, w_vacauous) for p in topic_word_dist.T]
    return vacauous_distances_kl


# Background Distribution (D-BGround)
def ts_bground(doc_topic_dist):
    n_documents = doc_topic_dist.shape[1]
    d_bground = np.ones(n_documents) / n_documents
    d_bground_distances_kl = [kl_divergence(p.T, d_bground) for p in doc_topic_dist]
    return d_bground_distances_kl


# SwitchP
def switchp(phi, texts):
    words = phi.index.to_list()
    max_topic_word_dist = np.argmax(phi.to_numpy(), axis=1)
    max_topic_word_dist = dict(zip(words, max_topic_word_dist))
    switchp_scores = []
    for text in texts:
        mapped_text = [
            max_topic_word_dist[word]
            for word in text.split()
            if word in max_topic_word_dist
        ]
        switches = (np.diff(mapped_text) != 0).sum()
        switchp_scores.append(switches / (len(mapped_text) - 1))

    return switchp_scores
